# If using installed version of asyncclick
#import asyncclick as click

# If using local copy of asyncclick
from . import asyncclick as click

from .click_colors import HelpColorsCommand, HelpColorsGroup


def group(name=None, **attrs):
    defaults = {
        'cls': HelpColorsGroup,
        'help_headers_color': 'yellow',
        'help_options_color': 'green',
    }
    kwargs = {**defaults, **attrs}
    return click.group(name=name, **kwargs)


def command(name=None, cls=None, **attrs):
    # Click cmd kargs to add colored output
    defaults = {
        'cls': HelpColorsCommand,
        'help_headers_color': 'yellow',
        'help_options_color': 'green',
    }
    kwargs = {**defaults, **attrs}
    return click.command(name=name, **kwargs)



# -*- coding: utf-8 -*-
# This module is based on the excellent work by Adam Bartoš who
# provided a lot of what went into the implementation here in
# the discussion to issue1602 in the Python bug tracker.
#
# There are some general differences in regards to how this works
# compared to the original patches as we do not need to patch
# the entire interpreter but just work in our little world of
# echo and prmopt.
import ctypes
import io
import os
import sys
import time
import zlib
from ctypes import byref
from ctypes import c_char
from ctypes import c_char_p
from ctypes import c_int
from ctypes import c_ssize_t
from ctypes import c_ulong
from ctypes import c_void_p
from ctypes import POINTER
from ctypes import py_object
from ctypes import windll
from ctypes import WinError
from ctypes import WINFUNCTYPE
from ctypes.wintypes import DWORD
from ctypes.wintypes import HANDLE
from ctypes.wintypes import LPCWSTR
from ctypes.wintypes import LPWSTR

import msvcrt

from ._compat import _NonClosingTextIOWrapper
from ._compat import text_type

try:
    from ctypes import pythonapi

    PyObject_GetBuffer = pythonapi.PyObject_GetBuffer
    PyBuffer_Release = pythonapi.PyBuffer_Release
except ImportError:
    pythonapi = None


c_ssize_p = POINTER(c_ssize_t)

kernel32 = windll.kernel32
GetStdHandle = kernel32.GetStdHandle
ReadConsoleW = kernel32.ReadConsoleW
WriteConsoleW = kernel32.WriteConsoleW
GetConsoleMode = kernel32.GetConsoleMode
GetLastError = kernel32.GetLastError
GetCommandLineW = WINFUNCTYPE(LPWSTR)(("GetCommandLineW", windll.kernel32))
CommandLineToArgvW = WINFUNCTYPE(POINTER(LPWSTR), LPCWSTR, POINTER(c_int))(
    ("CommandLineToArgvW", windll.shell32)
)
LocalFree = WINFUNCTYPE(ctypes.c_void_p, ctypes.c_void_p)(
    ("LocalFree", windll.kernel32)
)


STDIN_HANDLE = GetStdHandle(-10)
STDOUT_HANDLE = GetStdHandle(-11)
STDERR_HANDLE = GetStdHandle(-12)


PyBUF_SIMPLE = 0
PyBUF_WRITABLE = 1

ERROR_SUCCESS = 0
ERROR_NOT_ENOUGH_MEMORY = 8
ERROR_OPERATION_ABORTED = 995

STDIN_FILENO = 0
STDOUT_FILENO = 1
STDERR_FILENO = 2

EOF = b"\x1a"
MAX_BYTES_WRITTEN = 32767


class Py_buffer(ctypes.Structure):
    _fields_ = [
        ("buf", c_void_p),
        ("obj", py_object),
        ("len", c_ssize_t),
        ("itemsize", c_ssize_t),
        ("readonly", c_int),
        ("ndim", c_int),
        ("format", c_char_p),
        ("shape", c_ssize_p),
        ("strides", c_ssize_p),
        ("suboffsets", c_ssize_p),
        ("internal", c_void_p),
    ]


# On PyPy we cannot get buffers so our ability to operate here is
# serverly limited.
if pythonapi is None:
    get_buffer = None
else:

    def get_buffer(obj, writable=False):
        buf = Py_buffer()
        flags = PyBUF_WRITABLE if writable else PyBUF_SIMPLE
        PyObject_GetBuffer(py_object(obj), byref(buf), flags)
        try:
            buffer_type = c_char * buf.len
            return buffer_type.from_address(buf.buf)
        finally:
            PyBuffer_Release(byref(buf))


class _WindowsConsoleRawIOBase(io.RawIOBase):
    def __init__(self, handle):
        self.handle = handle

    def isatty(self):
        io.RawIOBase.isatty(self)
        return True


class _WindowsConsoleReader(_WindowsConsoleRawIOBase):
    def readable(self):
        return True

    def readinto(self, b):
        bytes_to_be_read = len(b)
        if not bytes_to_be_read:
            return 0
        elif bytes_to_be_read % 2:
            raise ValueError(
                "cannot read odd number of bytes from UTF-16-LE encoded console"
            )

        buffer = get_buffer(b, writable=True)
        code_units_to_be_read = bytes_to_be_read // 2
        code_units_read = c_ulong()

        rv = ReadConsoleW(
            HANDLE(self.handle),
            buffer,
            code_units_to_be_read,
            byref(code_units_read),
            None,
        )
        if GetLastError() == ERROR_OPERATION_ABORTED:
            # wait for KeyboardInterrupt
            time.sleep(0.1)
        if not rv:
            raise OSError("Windows error: {}".format(GetLastError()))

        if buffer[0] == EOF:
            return 0
        return 2 * code_units_read.value


class _WindowsConsoleWriter(_WindowsConsoleRawIOBase):
    def writable(self):
        return True

    @staticmethod
    def _get_error_message(errno):
        if errno == ERROR_SUCCESS:
            return "ERROR_SUCCESS"
        elif errno == ERROR_NOT_ENOUGH_MEMORY:
            return "ERROR_NOT_ENOUGH_MEMORY"
        return "Windows error {}".format(errno)

    def write(self, b):
        bytes_to_be_written = len(b)
        buf = get_buffer(b)
        code_units_to_be_written = min(bytes_to_be_written, MAX_BYTES_WRITTEN) // 2
        code_units_written = c_ulong()

        WriteConsoleW(
            HANDLE(self.handle),
            buf,
            code_units_to_be_written,
            byref(code_units_written),
            None,
        )
        bytes_written = 2 * code_units_written.value

        if bytes_written == 0 and bytes_to_be_written > 0:
            raise OSError(self._get_error_message(GetLastError()))
        return bytes_written


class ConsoleStream(object):
    def __init__(self, text_stream, byte_stream):
        self._text_stream = text_stream
        self.buffer = byte_stream

    @property
    def name(self):
        return self.buffer.name

    def write(self, x):
        if isinstance(x, text_type):
            return self._text_stream.write(x)
        try:
            self.flush()
        except Exception:
            pass
        return self.buffer.write(x)

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def __getattr__(self, name):
        return getattr(self._text_stream, name)

    def isatty(self):
        return self.buffer.isatty()

    def __repr__(self):
        return "<ConsoleStream name={!r} encoding={!r}>".format(
            self.name, self.encoding
        )


class WindowsChunkedWriter(object):
    """
    Wraps a stream (such as stdout), acting as a transparent proxy for all
    attribute access apart from method 'write()' which we wrap to write in
    limited chunks due to a Windows limitation on binary console streams.
    """

    def __init__(self, wrapped):
        # double-underscore everything to prevent clashes with names of
        # attributes on the wrapped stream object.
        self.__wrapped = wrapped

    def __getattr__(self, name):
        return getattr(self.__wrapped, name)

    def write(self, text):
        total_to_write = len(text)
        written = 0

        while written < total_to_write:
            to_write = min(total_to_write - written, MAX_BYTES_WRITTEN)
            self.__wrapped.write(text[written : written + to_write])
            written += to_write


def _get_text_stdin(buffer_stream):
    text_stream = _NonClosingTextIOWrapper(
        io.BufferedReader(_WindowsConsoleReader(STDIN_HANDLE)),
        "utf-16-le",
        "strict",
        line_buffering=True,
    )
    return ConsoleStream(text_stream, buffer_stream)


def _get_text_stdout(buffer_stream):
    text_stream = _NonClosingTextIOWrapper(
        io.BufferedWriter(_WindowsConsoleWriter(STDOUT_HANDLE)),
        "utf-16-le",
        "strict",
        line_buffering=True,
    )
    return ConsoleStream(text_stream, buffer_stream)


def _get_text_stderr(buffer_stream):
    text_stream = _NonClosingTextIOWrapper(
        io.BufferedWriter(_WindowsConsoleWriter(STDERR_HANDLE)),
        "utf-16-le",
        "strict",
        line_buffering=True,
    )
    return ConsoleStream(text_stream, buffer_stream)


_stream_factories = {
    0: _get_text_stdin,
    1: _get_text_stdout,
    2: _get_text_stderr,
}


def _is_console(f):
    if not hasattr(f, "fileno"):
        return False

    try:
        fileno = f.fileno()
    except OSError:
        return False

    handle = msvcrt.get_osfhandle(fileno)
    return bool(GetConsoleMode(handle, byref(DWORD())))


def _get_windows_console_stream(f, encoding, errors):
    if (
        get_buffer is not None
        and encoding in ("utf-16-le", None)
        and errors in ("strict", None)
        and _is_console(f)
    ):
        func = _stream_factories.get(f.fileno())
        if func is not None:
            f = getattr(f, 'buffer', None)
            if f is None:
                return None
            return func(f)



from .config import Config
from .package import Package
from .template import Template
from abc import ABC, abstractmethod
from uvicore.typing import Any, Dict, List, Tuple, OrderedDict, Union

try:
    from fastapi import FastAPI
    from starlette.applications import Starlette
except ImportError:  # pragma: nocover
    FastAPI = None  # type: ignore
    Starlette = None  # type: ignore

# Regular attributes (version: str = '0.0.0) vs
# @property @abstractmethod

# Defining attributes here does MAKE the attributes in the
# inheriting class implicitly.  But we don't really want that
# because the user now accept the abstrac classes defaults
# and is not forced to impliment them at all.  We want to FORCE
# the implimentation to define their own attribute values.

# So @property @abstractmethod is the best way to enforce
# requirements and conformity.

class Application(ABC):
    """Main uvicore application"""

    # Instance Variables
    @property
    @abstractmethod
    def version(self) -> str:
        """Uvicore framework version number"""
        pass

    @property
    @abstractmethod
    def debug(self) -> bool:
        """Debug mode for entire application"""
        pass

    @property
    @abstractmethod
    def perfs(self) -> List:
        """List of all perf dumps for performance tuning"""
        pass

    @property
    @abstractmethod
    def http(self) -> Union[Starlette, FastAPI]:
        """HTTP Server Instance"""
        pass

    # @property
    # @abstractmethod
    # def config(self) -> Config:
    #     """Configuration system"""
    #     pass

    @property
    @abstractmethod
    def providers(self) -> OrderedDict[str, Dict]:
        """OrderedDict of providers from all packages in proper dependency order"""
        pass

    @property
    @abstractmethod
    def registered(self) -> bool:
        """All providers have been registered"""
        pass

    @property
    @abstractmethod
    def booted(self) -> bool:
        """All providers have been booted"""
        pass

    @property
    @abstractmethod
    def is_console(self) -> bool:
        """App running from CLI (not serving web or API)"""
        pass

    @property
    @abstractmethod
    def is_http(self) -> bool:
        """App running as HTTP server (not as CLI)"""
        pass

    @property
    @abstractmethod
    def packages(self) -> Dict[str, Package]:
        """OrderedDict of all packages defined from providers"""
        pass

    @property
    @abstractmethod
    def path(self) -> str:
        """Base path of running application"""
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """Short name of running application"""
        pass

    @property
    @abstractmethod
    def main(self) -> str:
        """The main packages running this application"""
        pass

    @abstractmethod
    def bootstrap(self, app_config: Dict, path: str, is_console: bool) -> None:
        """Bootstrap the uvicore application"""
        pass

    @abstractmethod
    def package(self, package: str, *, name: str = None, main: bool = False) -> Package:
        """Get package by name or by main running package"""
        pass

    @abstractmethod
    def perf(self, item: any) -> None:
        """Add entry to debug performance counter"""
        pass

    # @abstractmethod
    # def dump(self, *args) -> None:
    #     """Pretty print args to console"""
    #     pass

    # @abstractmethod
    # def dd(self, *args) -> None:
    #     """Pretty print args to console and exit()"""
    #     pass



import uvicore
from xx_vendor.xx_appname.models import xx_ModelName
from uvicore.support.dumper import dump, dd


# ------------------------------------------------------------------------------
# Uvicore Database Seeder Schematic
# This schematic is filled with examples, a suppliment to the docs.
# Pick the best example for your use case and modify as needed!
# ------------------------------------------------------------------------------


@uvicore.seeder()
async def seed():
    uvicore.log.item('Seeding xx_tablename')


    # --------------------------------------------------------------------------
    # Example: Bulk insert List[Dict]
    # --------------------------------------------------------------------------
    await xx_ModelName.insert([
        {'slug': 'xx_modelname-1', 'title': 'xx_ModelName 1'},
        {'slug': 'xx_modelname-2', 'title': 'xx_ModelName 2'},
    ])


    # --------------------------------------------------------------------------
    # Example: Bulk insert List[xx_ModelName]
    # --------------------------------------------------------------------------
    # Issue here is some model fields are required that we don't have yet, like relations.
    # Also cannot add sub relations like one-to-many, many-to-many etc...
    # await xx_ModelName.insert([
    #     xx_ModelName(slug='xx_modelname-1', title='xx_ModelName 1'),
    #     xx_ModelName(slug='xx_modelname-2', title='xx_ModelName 2'),
    # ])


    # --------------------------------------------------------------------------
    # Example: Single insert (not bulk) using Model instance
    # --------------------------------------------------------------------------
    # tags = await Tag.query().key_by('name').get()
    # post = await xx_ModelName(slug='post-1', title='Post 1', creator_id=1).save()
    # # Create AND Link if not exist Many-To-Many tags
    # await post.link('tags', [
    #     tags['linux'],
    #     tags['bsd'],
    # ])
    # # Create Polymorphic One-To-One
    # await post.create('image', {
    #     'filename': 'post2-image.png',
    #     'size': 2483282
    # })
    # # Create Polymorphic One-To-Many
    # # NOTE: .add is simply an alias for .create()
    # await post.add('attributes', [
    #     {'key': 'post2-test1', 'value': 'value for post2-test1'},
    #     {'key': 'post2-test2', 'value': 'value for post2-test2'},
    #     {'key': 'badge', 'value': 'IT'},
    # ])



    # --------------------------------------------------------------------------
    # Example: Bulk insert List[Dict] with child relations
    # --------------------------------------------------------------------------
    # tags = await Tag.query().key_by('name').get()
    # await xx_ModelName.insert_with_relations([
    #     {
    #         'slug': 'post-1',
    #         'title': 'Post 1',
    #         'creator_id': 1,

    #         # One to many
    #         'comments': [
    #             {'title': 'Comment 1', 'body': 'Comment 1 body'},
    #             {'title': 'Comment 2', 'body': 'Comment 2 body'},
    #         ],

    #         # Many-To-Many tags works with existing Model, new Model or new Dict
    #         'tags': [
    #            # Existing Tag
    #            tags['linux'],
    #            tags['mac'],
    #            tags['bsd'],
    #            tags['bsd'],  # Yes its a duplicate, testing that it doesn't fail

    #            # New Tag as Model (tag created and linked)
    #            Tag(name='test1', creator_id=4),

    #            # New Tag as Dict (tag created and linked)
    #            {'name': 'test2', 'creator_id': 4},
    #         ],

    #         # Polymorphic One-To-One
    #         'image': {
    #             'filename': 'post1-image.png',
    #             'size': 1234932,
    #         },

    #         # Polymorphic One-To-Many Attributes
    #         'attributes': [
    #             {'key': 'post1-test1', 'value': 'value for post1-test1'},
    #             {'key': 'post1-test2', 'value': 'value for post1-test2'},
    #             {'key': 'badge', 'value': 'IT'},
    #         ],
    #     }
    # ])



    # --------------------------------------------------------------------------
    # Example: Faker with List[xx_ModelName]
    # --------------------------------------------------------------------------
    # from faker import Faker
    # xx_modelname_items = []
    # fake = Faker()
    # for _ in range(2):
    #     title = fake.text(max_nb_chars=50)
    #     xx_modelname = xx_ModelName(
    #         slug=fake.slug(title),
    #         title=title,
    #         creator_id=1,
    #     )
    #     xx_modelname_items.append(xx_modelname)
    # await xx_ModelName.insert(xx_modelname_items)



import uvicore
from uvicore.typing import Dict, List, OrderedDict, get_type_hints, Tuple
from uvicore.events import Handler
from uvicore.support import module
from uvicore.support.dumper import dump, dd
from uvicore.foundation.events.app import Booted as OnAppBooted
from uvicore.contracts import Package as Package
from uvicore.console import command_is
from starlette.applications import Starlette
from fastapi import FastAPI, Depends
from uvicore.http import response
from uvicore.http.events import server as HttpServerEvents
from uvicore.http.routing.router import Routes
from functools import partial, update_wrapper
from uvicore.http.routing import ApiRoute, WebRoute
from uvicore.http.routing import Guard
from uvicore.http.openapi.docs import get_swagger_ui_html, get_swagger_ui_oauth2_redirect_html
from uvicore.http.rapidoc.docs import get_rapidoc_ui_html, get_rapidoc_pdf_ui_html, get_rapidoc_ui_oauth2_redirect_html

# Temp speed testing variable to swap Web router from FastAPI to Starlette
# NOTICE if you set this to TRUE, all your type hinded /{vars} and ?vars will NOT
# work in your routes and you will have to revert to Starlettes way of
# handling vars with request.path_params['user_id'] which will break all your apps.
USE_STARLETTE=False


class Http(Handler):

    def __call__(self, event: OnAppBooted):
        """Bootstrap the HTTP Server after the Application is Booted"""

        # This will never fire if the uvicore.http module is not loaded
        # This will fire even if in console mode so we can handle showing routes
        # on certain commands.

        # Import each packages router files and add their routes to the package definition
        self.build_package_routes()

        # Merge all packages Web and Api routes
        (web_routes, api_routes) = self.merge_routes()

        # Fire up the HTTP server only if running from HTTP
        # Notice this is below building package routes above.  This is because
        # package routes are still built with certain CLI groups like package list and http route.
        # The server however is NOT fired up if not running a full http server command
        if not uvicore.app.is_http: return

        # If no routes, do nothing, no HTTP server at all
        if not web_routes and not api_routes: return

        # Fire up one or multiple HTTP servers.
        # If we have both web and api routes then we will mount subservers.
        # If not, we will only use one server.
        (base_server, web_server, api_server) = self.create_http_servers(web_routes, api_routes)

        #dd("Base: ", base_server, "WEB: ", web_server, "API: ", api_server)

        # Add global web and api specific middleware
        self.add_middleware(web_server, api_server)

        # Add global web and api specific exception handlers
        self.add_exception_handlers(web_server, api_server)

        # Get configured web and api prefixes
        web_prefix = self.get_prefix('app.web.prefix')
        api_prefix = self.get_prefix('app.api.prefix')

        # Add web routes to the web server
        self.add_web_routes(web_server, web_routes)

        # Add api routes to the api server
        self.add_api_routes(api_server, api_routes)

        # Add web paths (public, asset, view), configure templates and view composers
        self.configure_webserver(web_server)

        # Mount our sub servers into our base server
        if web_server and api_server:
            # Running both web and api servers, we must have a perent base server
            # to mount the web and api sub servers into.  This is how we properly
            # separate web vs api middleware and other configurations.
            # Longer prefix has to be FIRST to match subapp regex properly
            if len(web_prefix) > len(api_prefix):
                base_server.mount(web_prefix, web_server)  # Web first
                base_server.mount(api_prefix, api_server)
            else:
                base_server.mount(api_prefix, api_server)  # Api first
                base_server.mount(web_prefix, web_server)
        elif web_server:
            base_server.mount(web_prefix, web_server)
        elif api_server:
            base_server.mount(api_prefix, api_server)

        # Add our base server to uvicore
        uvicore.app._http = base_server


        # Attach to Starlette events and translate into Uvicore events
        @uvicore.app.http.on_event("startup")
        async def startup():
            await HttpServerEvents.Startup().dispatch_async()

        @uvicore.app.http.on_event("shutdown")
        async def shutdown():
            await HttpServerEvents.Shutdown().dispatch_async()

        # Debug and dump the actual HTTP servers (base, web, api) info and routes
        debug_dump = False
        if debug_dump:
            dump("Base: ", base_server, "WEB: ", web_server, "API: ", api_server)

            dump('#################################################################')
            dump("Main HTTP Server APPLICATION", uvicore.app.http.__dict__)

            dump('#################################################################')
            dump("Main HTTP Server Routes")
            for route in uvicore.app.http.routes:
                dump(route.__dict__)

            for route in uvicore.app.http.routes:
                if route.name is None:
                    dump('#################################################################')
                    dump("Sub Server '{}' APPLICATION".format(route.path))
                    dump(route.app.__dict__)

                    dump('#################################################################')
                    dump("Sub Server '{}' Routes".format(route.path))
                    for r in route.app.router.routes:
                        dump(r.__dict__)

    def build_package_routes(self) -> None:
        """Import all packages web and api routes files and add to packages route definition"""
        for package in uvicore.app.packages.values():
            if package.web.routes_module or package.api.routes_module:
                package.web.routes = self.import_package_web_routes(package)
                package.api.routes = self.import_package_api_routes(package)

    def import_package_web_routes(self, package: Package) -> Dict:
        """Import one package web routes and return routes Dict"""

        # If no routes_modules defined, nothing todo
        if not package.web.routes_module: return

        # Allow only if running as HTTP or from certain CLI commands like package list/show...
        if package.registers.web_routes and (
            uvicore.app.is_http or
            command_is('http') or
            command_is('package') or
            command_is('ioc') or
            command_is('app')
        ) == False: return

        routes_module = package.web.routes_module
        prefix = package.web.prefix
        name_prefix = package.web.name_prefix

        # Import main web routes module
        routes: Routes = module.load(routes_module).object(package)

        # Get name prefix from package name plus custom name prefix
        if name_prefix:
            if name_prefix[0] == '.': name_prefix = name_prefix[1:]
            name_prefix = package.short_name + '.' + name_prefix
        else:
            name_prefix = package.short_name

        # Add to application running config for debug overview only
        uvicore.app.add_running_config('http.web.route_modules.[' + routes_module + '].prefix', prefix)
        uvicore.app.add_running_config('http.web.route_modules.[' + routes_module + '].name_prefix', name_prefix)

        # Import the web router and create a new instance
        from uvicore.http.routing.web_router import WebRouter  # isort:skip
        router = WebRouter(package, prefix, name_prefix)

        # Get returned router with all defined routes
        router = routes.register(router)

        # Merge routes level middleware into each route
        routes_middleware = routes._middleware()
        if routes_middleware:
            for route in router.routes.values():
                (route.middleware, route.endpoint) = router._merge_route_middleware(routes_middleware, route.middleware, route.endpoint)

        # Return routes
        return router.routes

    def import_package_api_routes(self,  package: Package) -> Dict:
        """Import one package api routes and return routes Dict"""

        # If no routes_modules defined, nothing todo
        if not package.api.routes_module: return

        # Allow only if running as HTTP or from certain CLI commands like package list/show...
        if package.registers.api_routes and (
            uvicore.app.is_http or
            command_is('http') or
            command_is('package') or
            command_is('ioc') or
            command_is('app')
        ) == False: return

        routes_module = package.api.routes_module
        prefix = package.api.prefix
        name_prefix = package.api.name_prefix

        # Import main web routes module
        routes: Routes = module.load(routes_module).object(package)

        # Get name prefix from package name plus custom name prefix
        if name_prefix:
            if name_prefix[0] == '.': name_prefix = name_prefix[1:]
            name_prefix = package.short_name + '.' + name_prefix
        else:
            name_prefix = package.short_name

        # Add to application running config for debug overview only
        uvicore.app.add_running_config('http.api.route_modules.[' + routes_module + '].prefix', prefix)
        uvicore.app.add_running_config('http.api.route_modules.[' + routes_module + '].name_prefix', name_prefix)

        # Import the api router and create a new instance
        from uvicore.http.routing.api_router import ApiRouter  # isort:skip
        router = ApiRouter(package, prefix, name_prefix)

        # Get returned router with all defined routes
        router = routes.register(router)

        # Merge routes level middleware into each route
        routes_middleware = routes._middleware()
        if routes_middleware:
            for route in router.routes.values():
                (route.middleware, route.endpoint) = router._merge_route_middleware(routes_middleware, route.middleware, route.endpoint)

        #Return routes
        return router.routes

    def merge_routes(self) -> Tuple[Dict[str, WebRoute], Dict[str, ApiRoute]]:
        """Merge all packages web and api routes together"""

        # This merging means we can overwrite other packages routes.
        # Last package define wins in a route override battle
        web_routes = Dict()
        api_routes = Dict()
        package: Package
        for package in uvicore.app.packages.values():
            # Merging in package order ensures last package wins!
            if package.web.routes and package.registers.web_routes:
               web_routes.merge(package.web.routes)
            if package.api.routes and package.registers.api_routes:
                api_routes.merge(package.api.routes)

        # Add to application running config for debug overview only
        uvicore.app.add_running_config('http.web.routes', web_routes)
        uvicore.app.add_running_config('http.api.routes', api_routes)

        # Return routes tuple
        return (web_routes, api_routes)

    def create_http_servers(self, web_routes: Dict[str, WebRoute], api_routes: Dict[str, ApiRoute]) -> Tuple[Starlette, FastAPI, FastAPI]:
        """Fire up one or multiple HTTP servers"""

        # If we have both web and api routes then we will submount subservers.
        # If not, we will only use one server.
        base_server: Starlette = None
        web_server: FastAPI = None
        api_server: FastAPI = None
        debug = uvicore.config('app.debug')

        # We ALWAYS have 2 servers and the base_server (parent) is alway starlette
        # Why 2 servers?  Because of our web and api prefixes.  The web and api servers
        # get "mounted" inside the base wtih a special prefix to separate all routes and api docs

        # Base server is always starlette
        base_server = Starlette(debug=debug)

        # Web server is always a sub Starlette server
        if web_routes:
            if USE_STARLETTE:
                web_server = Starlette(debug=debug)
            else:
                web_server = FastAPI(
                    debug=debug,
                    version=uvicore.app.version,
                    openapi_url=None,
                    swagger_ui_oauth2_redirect_url=None,
                )

        # API server is always a sub FastAPI server
        if api_routes:
            api_server = FastAPI(
                debug=debug,

                #openapi_prefix=api_prefix,
                #root_path=api_prefix,
                #openapi_url='/api/openapi.json',

                #docs_url='/api/docs',
                #redoc_url='/api/`redoc`',

                # This title is what shows up in OpenAPI <h1>, not the HTML <title>, which is defined in add_api_routes def openapi_docs()
                title=uvicore.config('app.api.openapi.title'),

                # Get version from main running apps package.version
                version=uvicore.app.package(main=True).version,

                # NOTE: Most other info is provided in my add_api_routes def openapi_docs() override

                #openapi_url=uvicore.config('app.api.openapi.url'),
                #swagger_ui_oauth2_redirect_url=uvicore.config('app.api.openapi.docs_url') + '/oauth2-redirect',

                # Default FastAPI docs and redoc routes
                #docs_url=uvicore.config('app.api.openapi.docs_url'),
                #redoc_url=uvicore.config('app.api.openapi.redoc_url'),

                # DISABLE the /docs and /redoc URLs
                # Why?  Because we add our own CUSTOM one later in add_api_routes()
                docs_url=None,

                # DISABLE redoc
                # Can leave redoc on for now, maybe add to config to change /redoc path
                redoc_url=None,

                #swagger_ui_oauth2_redirect_url=

                #root_path='/api',  #fixme with trying out kong
            )



        ############ ORIGINAL BELOW

        # # Base server is Starlette
        # if web_routes and api_routes:
        #     base_server = Starlette(
        #         debug=debug,
        #     )

        # # Web server is FastAPI with NO OpenAPI setup
        # if web_routes:
        #     if USE_STARLETTE:
        #         # Try a pure starlette server
        #         web_server = Starlette(debug=debug)
        #     else:
        #         web_server = FastAPI(
        #             debug=debug,
        #             version=uvicore.app.version,
        #             openapi_url=None,
        #             swagger_ui_oauth2_redirect_url=None,
        #         )

        # # Api server is FastAPI with OpenAPI setup
        # if api_routes:
        #     api_server = FastAPI(
        #         debug=debug,

        #         # This title is what shows up in OpenAPI <h1>, not the HTML <title>, which is defined in add_api_routes def openapi_docs()
        #         title=uvicore.config('app.api.openapi.title'),

        #         # Get version from main running apps package.version
        #         version=uvicore.app.package(main=True).version,

        #         # NOTE: Most other info is provided in my add_api_routes def openapi_docs() override

        #         #openapi_url=uvicore.config('app.api.openapi.url'),
        #         #swagger_ui_oauth2_redirect_url=uvicore.config('app.api.openapi.docs_url') + '/oauth2-redirect',

        #         # Default FastAPI docs and redoc routes
        #         #docs_url=uvicore.config('app.api.openapi.docs_url'),
        #         #redoc_url=uvicore.config('app.api.openapi.redoc_url'),

        #         docs_url=None,
        #         redoc_url=None,

        #         #swagger_ui_oauth2_redirect_url=

        #         #root_path='/api',  #fixme with trying out kong
        #     )
        return (base_server, web_server, api_server)

    def add_middleware(self, web_server: FastAPI, api_server: FastAPI) -> None:
        """Add global web and api middleware to their respective servers"""

        # Add global web middleware
        if web_server:
            for name, middleware in uvicore.config.app.web.middleware.items():
                cls = module.load(middleware.module).object
                web_server.add_middleware(cls, **middleware.options)

        # Add global api middleware
        if api_server:
            for name, middleware in uvicore.config.app.api.middleware.items():
                cls = module.load(middleware.module).object
                api_server.add_middleware(cls, **middleware.options)

    def add_exception_handlers(self, web_server: FastAPI, api_server: FastAPI) -> None:
        """Add global web and api exception handlers to their respective servers"""
        from starlette.exceptions import HTTPException

        # Add global web exception handlers
        if web_server and uvicore.config.app.web.exception.handler:
            cls = module.load(uvicore.config.app.web.exception.handler).object
            web_server.add_exception_handler(HTTPException, cls)
            # for name, handler in uvicore.config.app.web.exceptions.items():
            #     if str(name).lower() == 'main':
            #         cls = module.load(handler).object
            #         web_server.add_exception_handler(HTTPException, cls)

        # Add global api exception handlers
        if api_server and uvicore.config.app.api.exception.handler:
            cls = module.load(uvicore.config.app.api.exception.handler).object
            api_server.add_exception_handler(HTTPException, cls)

            # for name, handler in uvicore.config.app.api.exceptions.items():
            #     if str(name).lower() == 'main':
            #         cls = module.load(handler).object
            #         api_server.add_exception_handler(HTTPException, cls)

        #dump(api_server.__dict__)
        #dump(web_server.__dict__)

    def add_web_routes(self, web_server, web_routes: Dict[str, WebRoute]) -> None:
        """Add web routes to the web server"""

        # Do nothing if no api_routes are defined
        if not web_routes: return

        for route in web_routes.values():
            if USE_STARLETTE:
                # Starlette
                web_server.add_route(
                    path=route.path,
                    route=route.endpoint,
                    methods=route.methods,
                    name=route.name,
                    include_in_schema=False,
                )
            else:
                # FastAPI
                web_server.add_api_route(
                    path=route.path or '/',
                    endpoint=route.endpoint,
                    methods=route.methods,
                    name=route.name,
                    include_in_schema=False,
                    response_class=response.HTML,
                    dependencies=route.middleware,
                )

    def add_api_routes(self, api_server, api_routes: Dict[str, ApiRoute]) -> None:
        """Add api routes to the api server"""

        # Do nothing if no api_routes are defined
        if not api_routes: return

        # Get important app configs
        api_prefix = self.get_prefix('app.api.prefix')
        openapi = uvicore.config.app.api.openapi
        rapidoc = uvicore.config.app.api.rapidoc
        oauth2 = uvicore.config.app.auth.oauth2

        # Determine if OpenAPI docs oauth2 authentication is enabled
        # If OpenAPI oauth2 is enabled, create a FastAPI OAuth2AuthorizationCodeBearer variable
        if openapi.oauth2_enabled and oauth2:
            from fastapi.security import OAuth2AuthorizationCodeBearer
            oauth2_scheme = OAuth2AuthorizationCodeBearer(
                authorizationUrl=oauth2.base_url + oauth2.authorize_path,
                tokenUrl=oauth2.base_url + oauth2.token_path,

                # This little flag here is CRITICAL it be false.  If true
                # then fastapi/security/oauth2.py OAuth2AuthorizationCodeBearer calss __call__() method
                # will throw a '401 Not authenticated' if there is no token.  But Uvicore allows anonymous
                # endpoints EVEN when that endpoint requires permission scopes.  This is because a non logged in user (no token, invalid token)
                # is tied to the Anonymous user which still has roles and permissions which our endpoint scopes still obey.
                # So setting this to False means FastAPI will NOT throw the 401 for us and let Uvicore handle scope validation
                # in uvicore/http/routing/guard.py even if you are anonymous!
                auto_error=False,
            )

        # Create our own custom RapiDoc docs route
        if rapidoc.path and openapi.path:

            # Oauth2 redirect URL (without api_prefid)
            rapidoc_redirect_url =  rapidoc.path + '/oauth-receiver.html'

            @api_server.get(rapidoc.path, include_in_schema=False)
            def rapidoc_docs():
                return get_rapidoc_ui_html(
                    openapi_url=api_prefix + openapi.path,
                )

            @api_server.get(rapidoc.pdf_path, include_in_schema=False)
            def rapidoc_docs():
                return get_rapidoc_pdf_ui_html(
                    openapi_url=api_prefix + openapi.path,
                )

            @api_server.get(rapidoc_redirect_url, include_in_schema=False)
            def rapidoc_redirect():
                return get_rapidoc_ui_oauth2_redirect_html()


        # Create our own custom OpenAPI docs route
        if openapi.path and openapi.docs.path:

            # Oauth2 redirect URL (without api_prefid)
            openapi_redirect_url =  openapi.docs.path + '/login'

            @api_server.get(openapi.docs.path, include_in_schema=False)
            def openapi_docs():
                return get_swagger_ui_html(

                    # Define our openapi.json path from prefix and config
                    openapi_url=api_prefix + openapi.path,
                    oauth2_redirect_url=api_prefix + openapi_redirect_url,

                    # This title is html <title> while the title in app_server = FastAPI is what shows up in OpenAPI <h1> title, I make them the same
                    title=openapi.title,

                    # Swagger urls
                    swagger_favicon_url=openapi.docs.favicon_url,
                    swagger_js_url=openapi.docs.js_url,
                    swagger_css_url=openapi.docs.css_url,

                    init_oauth={
                        'clientId': oauth2.client_id,
                        #'clientSecret': "GaMz_F83_KB8ac6g-Eds0uoHyeoxg03X184yBqZR5Ws",
                        #'realm': "https://auth-local.triglobal.io",
                        'appName': uvicore.config.app.name,
                        #'scopeSeparator': " ",
                        'scopes': "openid profile",
                        #'additionalQueryStringParams': {'client_id': "7cc7d2a5-cc02-43ca-93bc-8476370ebf9d"},
                        #'usePkceWithAuthorizationCodeGrant': False
                    },
                    doc_expansion=openapi.docs.expansion.lower() or 'list'
                )

            @api_server.get(openapi_redirect_url, include_in_schema=False)
            def openapi_redirect():
                return get_swagger_ui_oauth2_redirect_html()


        # Loop each uvicore route and add as FastAPI route
        for route in api_routes.values():
            # ----------------------------------------------------------------------------------
            # NO, all of this moved into api_router.py so I can see it at the package definition (./uvicore package list) level
            # ----------------------------------------------------------------------------------
            # endpoint_func = route.endpoint
            # response_model = route.response_model

            # # If endpoint is partial, grab inside func for type hindint and docstrings
            # if isinstance(route.endpoint, partial):
            #     # Endpoint is a partial (was overwritten to default some higher order middleware)
            #     # A partial overwrites the original docstring.  Functools update_wrapper will copy it back
            #     # as well as handle merging of other important properties
            #     #update_wrapper(route.endpoint, route.endpoint.func)
            #     endpoint_func = route.endpoint.func

            #     # Blank out the __doc__ on actual Partial itself, not actual endpoint inside partial.
            #     # If not, OpenAPI doc description will be partial(func, *args, **keywords) - new function with partial application etc...
            #     route.endpoint.__doc__ = None

            # # Get response model from parameter or infer from endpoint return type hint
            # #response_model = route.response_model if route.response_model else get_type_hints(endpoint_func).get('return')
            # response_model = route.response_model or get_type_hints(endpoint_func).get('return')

            # # Get openapi description from route param or endpoint docstring
            # description = route.description or endpoint_func.__doc__
            # ----------------------------------------------------------------------------------

            # If OpenAPI oauth2 authentication is enabled, add the proper route
            # dependency to our oauth2_schema
            if openapi.oauth2_enabled:
                found_guard = False
                for middleware in route.middleware:
                    if type(middleware) == Guard:
                        found_guard = True
                        break;

                if found_guard:
                    route.middleware.append(Depends(oauth2_scheme))

            # Add uvicore route to FastAPI route
            api_server.add_api_route(
                #path=(prefix + route.path) or '/',
                path=(route.path) or '/',
                endpoint=route.endpoint,
                methods=route.methods,
                name=route.name,
                response_model=route.response_model,
                response_class=route.response_class or response.JSONResponse,
                responses=route.responses,
                tags=route.tags,
                dependencies=route.middleware,
                summary=route.summary,
                description=route.description,
            )

    def configure_webserver(self, web_server) -> None:
        """Configure the webserver with public, assets, views, composers and template options"""

        # Ignore if web_server was never fired up
        if not web_server: return

        # Loop each package with an HTTP definition and add to our HTTP server
        public_paths = []
        asset_paths = []
        view_paths = []
        view_composers = Dict()
        context_processors = Dict()
        for package in uvicore.app.packages.values():
            #if not 'http' in package: continue
            if not 'web' in package: continue

            # Append public paths for later
            for path in package.web.public_paths:
                path_location = module.location(path)
                if path_location not in public_paths: public_paths.append(path_location)

            # Append asset paths for later
            for path in package.web.asset_paths:
                path_location = module.location(path)
                if path_location not in asset_paths: asset_paths.append(path_location)

            # Append template paths
            for path in package.web.view_paths or []:
                view_paths.append(path)

            # Deep merge view composers
            view_composers.merge(package.web.view_composers or {})

            # Deep merge template options
            context_processors.merge(package.web.context_processors or {})

        # Save merged view composers into our config for use later (in http/response View() )
        uvicore.config.uvicore.http.view_composers = view_composers

        #dd(view_composers)

        # Mount all static paths
        self.mount_static(web_server, public_paths, asset_paths)

        # Initialize new template environment from the options above
        self.initialize_templates(view_paths, context_processors)

    def mount_static(self, web_server, public_paths: List, asset_paths: List) -> None:
        """Mount static routes defined in all packages"""
        StaticFiles = uvicore.ioc.make('uvicore.http.static.StaticFiles')

        # Mount all asset paths
        # Last directory defined WINS, which fits our last provider wins
        asset_url = uvicore.config.app.web.asset.path or '/assets'
        web_server.mount(asset_url, StaticFiles(directories=asset_paths), name='assets')

        # Mount all public paths (always at /)
        # Since it is root / MUST be defined last or it wins above any path after it
        web_server.mount('/', StaticFiles(directories=public_paths), name='public')

    def initialize_templates(self, paths, context_processors) -> None:
        """Initialize and configure template system"""
        # Get the template singleton from the IoC
        templates = uvicore.ioc.make('uvicore.templating.engine.Templates')

        # Add all packages view paths
        # First path wins, so we must REVERSE the package order
        # This will set the main running app FIRST as we always want the app to win
        paths.reverse()
        for path in paths:
            templates.include_path(module.location(path))

        # Add all packages deep_merged template context_processors
        if 'context_functions' in context_processors:
            for name, method in context_processors['context_functions'].items():
                templates.include_context_function(name, method)
        if 'context_filters' in context_processors:
            for name, method in context_processors['context_filters'].items():
                templates.include_context_filter(name, method)
        if 'filters' in context_processors:
            for name, method in context_processors['filters'].items():
                templates.include_filter(name, method)
        if 'tests' in context_processors:
            for name, method in context_processors['tests'].items():
                templates.include_test(name, method)

        # Initialize template system
        templates._init()

    def get_prefix(self, confpath: str) -> str:
        """Get configured web and api prefixes ensuring "" prefix and no trailing /, or / if blank)"""
        prefix = uvicore.config(confpath)
        #if not prefix: prefix = '/'
        #if prefix != '/' and prefix[-1] == '/': prefix = prefix[0:-1]  # No trailing /
        if prefix and prefix[-1] == '/': prefix = prefix[0:-1]  # No trailing /
        return prefix



from uvicore.typing import Dict, List
from uvicore.database import Connection
from uvicore.support.dumper import dump, dd
from uvicore.support.module import location


class Db:
    """Database Service Provider Mixin"""

    # def _add_db_definition(self, key, value):
        # if type(value) == list:
        #     if not self.package.database[key]: self.package.database[key] = []
        #     self.package.database[key].extend(value)
        # elif type(value) == dict:
        #     self.package.database[key].merge(value)
        # else:
        #     self.package.database[key] = value

        #self.package.database[key] = value

        # if 'database' not in self.package:
        #     self.package.database = Dict()
        # if type(value) == list:
        #     if key not in self.package.database: self.package.database = []
        #     self.package['database'][key].extend(value)
        # else:
        #     self.package['database'][key] = value

    def register_db_connections(self, connections: Dict, default: str):
        # for name, connection in connections.items():

        #     # Build URL and metakey
        #     # Metakey cannot be the connection name.  If 2 connections share the exact
        #     # same database (host, port, dbname) then they need to also share the same
        #     # metedata for foreign keys to work properly.

        #     if not connection.backend: connection.backend = 'sqlalchemy'
        #     connection.backend = connection.backend.lower()
        #     connection.driver = connection.driver.lower()
        #     url = ''
        #     metakey = ''

        #     if connection.backend == 'sqlalchemy':
        #         if connection.dialect == 'sqlite':
        #             url = 'sqlite:///' + connection.database
        #             metakey = url
        #         elif connection.dialect in ['mysql', 'postgresql', 'postgres']:
        #             url = (connection.dialect
        #                 + '+' + connection.driver
        #                 + '://' + connection.username
        #                 + ':' + connection.password
        #                 + '@' + connection.host
        #                 + ':' + str(connection.port)
        #                 + '/' + connection.database
        #             )
        #             metakey = (
        #                 connection.dialect
        #                 + '@' + connection.host
        #                 + ':' + str(connection.port)
        #                 + '/' + connection.database
        #             )
        #         # elif connection.dialect == 'snowflake':

        #         # Snowflake wip
        #         #from cryptography.hazmat.backends import default_backend
        #         #from cryptography.hazmat.primitives import serialization

        #         #     # NOTICE: FIXME
        #         #     # Snowflake does not work with database/encode
        #         #     # It may however work fine with SQLAlechemy 2.0 once we convert over

        #         #     # Once you get it working, here is what the CONFIG might look like
        #         #     # 'driver': 'snowflake',
        #         #     # # For private key auth - user, account and private_key are required (password should be blank)
        #         #     # 'account': env('DB_SF_ACCOUNT', 'account'),
        #         #     # 'database': env('DB_SF_DATABASE', ''),
        #         #     # 'schema': env('DB_SF_SCHEMA', ''),
        #         #     # 'warehouse': env('DB_SF_WAREHOUSE', ''),
        #         #     # 'role': env('DB_SF_ROLE', ''),
        #         #     # #'numpy': True,
        #         #     # #'cache_column_metadata': True,
        #         #     # # Username based
        #         #     # 'username': env('DB_SF_USERNAME', ''),
        #         #     # 'password': env('DB_SF_PASSWORD', ''),
        #         #     # #'authenticator': env('DB_SF_AUTHENTICATOR', 'externalbrowser'),
        #         #     # #'autocommit': True,
        #         #     # #'timezone': env('TIMEZONE', 'America/Chicago'),
        #         #     # #'paramstyle': env('DB_SF_PARAM_STYLE', 'numeric'),
        #         #     # 'private_key': env('DB_SF_PRIVATE_KEY', 'xxx'),
        #         #     # 'prefix': env('DB_SF_PREFIX', None),

        #         #     # See https://docs.snowflake.com/en/developer-guide/python-connector/sqlalchemy
        #         #     sep = '?'
        #         #     url += (
        #         #         'snowflake'
        #         #         + '://' + connection.username
        #         #         + ':' + connection.password
        #         #         + '@' + connection.account
        #         #     )

        #         #     # Optionally specify /database/schema
        #         #     if connection.database: url += '/' + connection.database
        #         #     if connection.schema: url += '/' + connection.schema

        #         #     # Using private key authentication
        #         #     # Required: username, account, private_key
        #         #     if connection.private_key:
        #         #         p_key = serialization.load_pem_private_key(
        #         #             (connection.private_key).encode('ascii'),
        #         #             password=None,
        #         #             backend=default_backend()
        #         #         )
        #         #         pkb = p_key.private_bytes(
        #         #             encoding=serialization.Encoding.DER,
        #         #             format=serialization.PrivateFormat.PKCS8,
        #         #             encryption_algorithm=serialization.NoEncryption()
        #         #         )
        #         #         url += "?connect_args={'private_key': " + str(pkb) + "}"
        #         #         sep = '&'

        #         #     if connection.warehouse:
        #         #         url += sep + 'warehouse=' + connection.warehouse
        #         #         sep = '&'
        #         #     if connection.role:
        #         #         url += sep + 'role=' + connection.role
        #         #         sep = '&'
        #         #     if connection.numpy:
        #         #         url += sep + 'numpy=' + str(connection.numpy)
        #         #         sep = '&'
        #         #     if connection.cache_column_metadata:
        #         #         url += sep + 'cache_column_metadata=' + str(connection.cache_column_metadata)
        #         #         sep = '&'

        #         #     metakey = (
        #         #         'snowflake'
        #         #         + '@' + connection.account
        #         #     )

        #     else:
        #         raise Exception(f"Connection backend {connection.backend} not supported")


        #     # Merge new values into connection SuperDict
        #     if not connection.prefix: connection.prefix = ''
        #     connection.merge({
        #         'name': name,
        #         'metakey': metakey.lower(),
        #         'url': url
        #     })
        #     #dd(connection, 'xx')

        self.package.database.connections = connections
        self.package.database.connection_default = default

    def register_db_models(self, items: List):
        # Default registration
        self.package.registers.defaults({'models': True})

        # Register models only if allowed
        if self.package.registers.models:
            self.package.database.models = items

    def register_db_tables(self, items: List):
        # Default registration
        self.package.registers.defaults({'tables': True})

        # Register tables only if allowed
        if self.package.registers.tables:
            self.package.database.tables = items

    def register_db_seeders(self, items: List):
        # Default registration
        self.package.registers.defaults({'seeders': True})

        # Register seeders only if allowed
        if self.package.registers.seeders:
            self.package.database.seeders = items



from uvicore.typing import Dict


class Cli:
    """CLI Service Provider Mixin"""

    def register_cli_commands(self, items: Dict[str, Dict] = None, *, group: str = None, help: str = None, commands: Dict = None):
        """Add commands as a dictionary or kwargs"""

        # Default registration
        self.package.registers.defaults({'commands': True})

        # Register commands only if allowed
        if not self.package.registers.commands: return

        if items:
            # Add as dictionary
            self.package.console.groups.merge(items)
        else:
            # Add as kwargs
            if help:
                self.package.console.groups.merge({
                    group: {
                        'help': help,
                        'commands': commands
                    }
                })
            else:
                self.package.console.groups.merge({
                    group: {
                        'commands': commands
                    }
                })



import uvicore
from uvicore.package import Provider
from uvicore.support.dumper import dump, dd
from uvicore.templating.package import bootstrap
from uvicore.foundation.events import app as AppEvents


@uvicore.provider()
class Templating(Provider):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        # Register event listeners
        # After all providers are booted we have a complete list of view paths
        # and template options fully merged.  Now we can fire up the static
        # paths and template system.
        AppEvents.Booted.listen(bootstrap.Templating)


    def boot(self) -> None:
        """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted, self.booted')"""

        # Define Provider Registrations
        self.registers(self.package.config.registers)




from uvicore.typing import Dict


class Redis:
    """Redis Service Provider Mixin"""

    def register_redis_connections(self, connections: Dict, default: str):
        self.package.redis.connections = connections
        self.package.redis.connection_default = default




# type: ignore
from .request import Request
from starlette.background import BackgroundTask, BackgroundTasks

# No, causes to much to be imported just by importing anything from uvicore.http
# from .routing import (ApiRoute, ApiRouter, Controller, ModelRouter, Router,
#                       Routes, WebRoute, WebRouter)


# NO - not used by user
#from .server import Server
# static.StaticFiles

# Do not import response, we want to import it as a file to use as a namespace
# from uvicore.http import response
# response.View() or response.Text().... so its a namespace

# The http package uses a __init__.py because users will import these
# methods from their own controllers and we want a nicer import which looks
# like - from uvicore.http import Request



import inspect
from typing import Any, Callable, List, Type, TypeVar, Union, get_type_hints

from fastapi import APIRouter, Depends
from pydantic.typing import is_classvar
from starlette.routing import Route, WebSocketRoute

"""
The MIT License (MIT)

Copyright (c) 2020 David Montague

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

T = TypeVar("T")

CBV_CLASS_KEY = "__cbv_class__"


def APIController(router: APIRouter) -> Callable[[Type[T]], Type[T]]:
    """
    This function returns a decorator that converts the decorated into a class-based view for the provided router.
    Any methods of the decorated class that are decorated as endpoints using the router provided to this function
    will become endpoints in the router. The first positional argument to the methods (typically `self`)
    will be populated with an instance created using FastAPI's dependency-injection.
    For more detail, review the documentation at
    https://fastapi-utils.davidmontague.xyz/user-guide/class-based-views/#the-cbv-decorator
    """

    def decorator(cls: Type[T]) -> Type[T]:
        return _cbv(router, cls)

    return decorator


def _cbv(router: APIRouter, cls: Type[T]) -> Type[T]:
    """
    Replaces any methods of the provided class `cls` that are endpoints of routes in `router` with updated
    function calls that will properly inject an instance of `cls`.
    """
    _init_cbv(cls)
    cbv_router = APIRouter()
    function_members = inspect.getmembers(cls, inspect.isfunction)
    functions_set = set(func for _, func in function_members)
    cbv_routes = [
        route
        for route in router.routes
        if isinstance(route, (Route, WebSocketRoute)) and route.endpoint in functions_set
    ]
    for route in cbv_routes:
        router.routes.remove(route)
        _update_cbv_route_endpoint_signature(cls, route)
        cbv_router.routes.append(route)
    router.include_router(cbv_router)
    return cls


def _init_cbv(cls: Type[Any]) -> None:
    """
    Idempotently modifies the provided `cls`, performing the following modifications:
    * The `__init__` function is updated to set any class-annotated dependencies as instance attributes
    * The `__signature__` attribute is updated to indicate to FastAPI what arguments should be passed to the initializer
    """
    if getattr(cls, CBV_CLASS_KEY, False):  # pragma: no cover
        return  # Already initialized
    old_init: Callable[..., Any] = cls.__init__
    old_signature = inspect.signature(old_init)
    old_parameters = list(old_signature.parameters.values())[1:]  # drop `self` parameter
    new_parameters = [
        x for x in old_parameters if x.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
    ]
    dependency_names: List[str] = []
    for name, hint in get_type_hints(cls).items():
        if is_classvar(hint):
            continue
        parameter_kwargs = {"default": getattr(cls, name, Ellipsis)}
        dependency_names.append(name)
        new_parameters.append(
            inspect.Parameter(name=name, kind=inspect.Parameter.KEYWORD_ONLY, annotation=hint, **parameter_kwargs)
        )
    new_signature = old_signature.replace(parameters=new_parameters)

    def new_init(self: Any, *args: Any, **kwargs: Any) -> None:
        for dep_name in dependency_names:
            dep_value = kwargs.pop(dep_name)
            setattr(self, dep_name, dep_value)
        old_init(self, *args, **kwargs)

    setattr(cls, "__signature__", new_signature)
    setattr(cls, "__init__", new_init)
    setattr(cls, CBV_CLASS_KEY, True)


def _update_cbv_route_endpoint_signature(cls: Type[Any], route: Union[Route, WebSocketRoute]) -> None:
    """
    Fixes the endpoint signature for a cbv route to ensure FastAPI performs dependency injection properly.
    """
    old_endpoint = route.endpoint
    old_signature = inspect.signature(old_endpoint)
    old_parameters: List[inspect.Parameter] = list(old_signature.parameters.values())
    old_first_parameter = old_parameters[0]
    new_first_parameter = old_first_parameter.replace(default=Depends(cls))
    new_parameters = [new_first_parameter] + [
        parameter.replace(kind=inspect.Parameter.KEYWORD_ONLY) for parameter in old_parameters[1:]
    ]
    new_signature = old_signature.replace(parameters=new_parameters)
    setattr(route.endpoint, "__signature__", new_signature)



import uvicore
import sqlalchemy as sa
from uvicore.contracts import Connection
from uvicore.support.dumper import dd, dump
from uvicore.contracts import Package as Package
from uvicore.database.query import DbQueryBuilder
from sqlalchemy.ext.asyncio import create_async_engine
from uvicore.contracts import Database as DatabaseInterface
from uvicore.typing import Any, Dict, List, Sequence, Mapping, Optional


@uvicore.service('uvicore.database.db.Db',
    aliases=['Database', 'database', 'db'],
    singleton=True,
)
class Db(DatabaseInterface):
    """Database private class.

    Do not import from this location.
    Use the uvicore.db singleton global instead."""

    # Backends like sqlalchemy
    SUPPORTED_BACKENDS = [
        'sqlalchemy'
    ]

    # Dialects like postgresql (or postgres), mysql, sqlite
    SUPPORTED_DIALECTS = [
        'postgresql',
        'postgres',
        'mysql',
        'sqlite',
    ]

    # Async drivers like
    # Drivers are like pymysql, aiomysql, asyncio, psycopg2...
    # This is limited to encode supported drivers
    SUPPORTED_SYNC_DRIVERS = [
        # MySQL
        'mysqldb',
        'pymysql',
        'mysqlconnector',
        'mariadbconnector',
        # Postgres
        'psycopg2',
        'pg8000',
        # SQLite
        'pysqlite',
        # Oracle
        'cx_oracle',
        # MSSQL
        'pyodbc',
        'pymssql',
    ]

    SUPPORTED_ASYNC_DRIVERS = [
        # MySQL
        'aiomysql',
        'asyncmy',
        # Postgres
        'asyncpg',
        'aiopg',
        # SQLite
        'aiosqlite'
    ]

    @property
    def default(self) -> str:
        """The default connection str for the main running app"""
        return self._default

    @property
    def connections(self) -> Dict[str, Connection]:
        """All connections from all packages, keyed by connection str name"""
        return self._connections

    @property
    def engines(self) -> Dict[str, sa.engine.Engine]:
        """All engines for all unique (by metakey) connections, keyed by metakey"""
        return self._engines

    @property
    def metadatas(self) -> Dict[str, sa.MetaData]:
        """All SQLAlchemy Metadata for all unique (by metakey) connections, keyed by metakey"""
        return self._metadatas

    def __init__(self) -> None:
        self._default = None
        self._connections = Dict()
        self._engines = Dict()
        self._metadatas = Dict()

    def init(self, default: str, connections: Dict[str, Connection]) -> None:
        """Initialize the database system with a default connection str and List of all Connections from all packages"""

        # Loop all connections from all packages
        connection: Connection
        for connection_name, connection in connections.items():

            # Define some defaults regardless of backend type
            connection.defaults({
                'name': connection_name,
                'backend': 'sqlalchemy',
                'dialect': 'sqlite',
            })

            # Standardize case
            connection.backend = connection.backend.lower()
            connection.dialect = connection.dialect.lower()

            # Validate supported backends
            if connection.backend not in self.SUPPORTED_BACKENDS:
                raise Exception(f"A packages config/database.py connection backend {connection.backend} not supported by Uvicore.  Must be one of [{','.join(self.SUPPORTED_BACKENDS)}].")

            # Validate supported dialects
            if connection.dialect not in self.SUPPORTED_DIALECTS:
                raise Exception(f"A packages config/database.py connection dialect {connection.dialect} not supported by Uvicore.  Must be one of [{','.join(self.SUPPORTED_DIALECTS)}].")

            # Build url and metakey from connection configuration
            if connection.backend == 'sqlalchemy':

                if connection.dialect in ['mysql', 'postgres', 'postgresql']:

                    # MySQL dialect defaults
                    if connection.dialect == 'mysql':
                        connection.defaults({
                            'dialect': 'mysql',
                            'driver': 'aiomysql',
                            'host': '127.0.0.1',
                            'port': 3306,
                            'database': 'mysql',
                            'username': '',
                            'password': '',
                            'prefix': None
                        })

                    # Postgres dialect defaults
                    else:
                        connection.defaults({
                            'dialect': 'postgres',
                            'driver': 'asyncpg',
                            'host': '127.0.0.1',
                            'port': 5432,
                            'database': 'postgres',
                            'username': '',
                            'password': '',
                            'prefix': None
                        })

                    # Build an SQLAlchemy compatible URL from connection configuration dictionary
                    # dialect+driver://<user>:<password>@<host>[:<port>]/<dbname>
                    conn_url = connection.url
                    if not conn_url:
                        conn_url = (sa.engine.url.URL.create(
                            drivername=str(connection.dialect) + '+' + str(connection.driver),
                            username=connection.username,
                            password=connection.password,
                            host=connection.host,
                            port=int(connection.port),
                            database=connection.database,
                        ))

                    # Build metakey
                    # Metakey is slightly different than the URL because we are trying to deduce
                    # a single SERVER/HOST, not including the separate database itself
                    connection.metakey = (
                        connection.dialect +
                        '@' + connection.host +
                        ':' + str(connection.port) +
                        '/' + connection.database
                    )

                # SQLite dialect defaults
                elif connection.dialect == 'sqlite':
                    connection.defaults({
                        'dialect': 'sqlite',
                        'driver': 'aiosqlite',
                        'host': '',
                        'port': '',
                        'database': ':memory:',
                        'prefix': None
                    })

                    # SQLite has a different URL
                    conn_url = connection.url
                    if not conn_url:
                        conn_url = (sa.engine.url.URL.create(
                            drivername=str(connection.dialect) + '+' + str(connection.driver),
                            database=connection.database,
                        ))

                    # SQLite has a different metakey
                    connection.metakey = connection.dialect + '://' + connection.database

                # Save conn_url as string
                connection.url = str(conn_url)

                # Attempt an async connection, else a sync connection
                # And automatically set an is_async attribute on our connection Dict
                try:
                    engine = create_async_engine(conn_url, connect_args=connection.options)
                    connection.is_async = True
                except:
                    engine = sa.create_engine(conn_url, connect_args=connection.options, pool_pre_ping=True)
                    connection.is_async = False

                # Add this new [sync or async] engine to our Dict of engines
                self._engines[connection.metakey] = engine

                # Add this new metadata to our Dict of metadatas
                self._metadatas[connection.metakey] = sa.MetaData()

        # Set instance variables
        self._default = default
        self._connections = connections

    def packages(self, connection: str = None, metakey: str = None) -> List[Package]:
        """Get all packages with the metakey (direct or derived from connection str)."""

        if not metakey:
            if not connection: connection = self.default
            metakey = self.connection(connection).metakey
        packages = []
        for package in uvicore.app.packages.values():
            if not 'database' in package: continue
            for conn in package.database.connections.values():
                if conn.metakey == metakey:
                    packages.append(package)
        return packages

    def metakey(self, connection: str = None, metakey: str = None) -> str:
        """Get one metekay by connection str or metakey"""
        try:
            if not metakey:
                if not connection:
                    connection = self.default
                metakey = self.connection(connection).metakey
            return metakey
        except Exception:
            dump("ERROR Connections:", self.connections)
            raise Exception('Metakey not found, connection={} metakey={}'.format(connection, metakey))

    def connection(self, connection: str = None) -> Connection:
        """Get one connection by connection name"""
        if not connection: connection = self.default
        return self.connections.get(connection)

    def metadata(self, connection: str = None, metakey: str = None) -> sa.MetaData:
        """Get one SQLAlchemy Metadata by connection str or metakey"""
        metakey = self.metakey(connection, metakey)
        return self.metadatas.get(metakey)

    def tables(self, connection: str = None, metakey: str = None) -> List[sa.Table]:
        """Get all SQLAlchemy tables for a given connection str or metakey"""
        metadata = self.metadata(connection, metakey)
        return metadata.tables

    def table(self, table: str, connection: str = None) -> sa.Table:
        """Get one SQLAlchemy Table by name (without prefix) and connection str or connection.tablename dot notation"""
        tablename = self.tablename(table, connection)
        metadata = self.metadata(connection)
        if metadata: return metadata.tables.get(tablename)

    def tablename(self, table: str, connection: str = None) -> str:
        """Get a SQLAlchemy tablename with prefix by name (without prefix) and connection str or connection.tablename dot notation"""
        if '.' in table:
            connection, table = tuple(table.split('.'))
        connection = self.connection(connection)
        if connection:
            if connection.prefix: return connection.prefix + table
            return table

    def engine(self, connection: str = None, metakey: str = None) -> sa.engine.Engine:
        """Get one SQLAlchemy Engine by connection str or metakey"""
        metakey = self.metakey(connection, metakey)
        return self.engines.get(metakey)

    def query(self, connection: str = None) -> DbQueryBuilder[DbQueryBuilder, Any]:
        """Database query builder passthrough"""
        if not connection: connection = self.default
        return DbQueryBuilder(connection)

    async def execute(
        self,
        query: Any,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.CursorResult:
        """Execute a SQLAlchemy Core Query based on connection str or metakey"""

        # Get engine for this (or default) connection
        engine = self.engine(connection, metakey)

        # Convert connection string into actual connection Dict
        connection: Connection = self.connection(connection)

        # If using raw SQL string
        if type(query) == str: query = sa.text(query)

        conn: sa.Connection
        if connection.is_async:
            # Execute with async driver
            async with engine.begin() as conn:
                # To dump raw SQL use str() - dd(str(query))
                # To get single inserted PK - result.inserted_primary_key
                # To get bulk inserted PK lists (not supported by MySQL) - result.inserted_primary_key_rows
                result: sa.CursorResult = await conn.execute(query, values)
                return result
        else:
            with engine.begin() as conn:
                result: sa.CursorResult = conn.execute(query, values)
                return result

    async def all(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Sequence[sa.Row]:
        """Get many records from query. Returns empty List if no records found"""
        result = await self.execute(query, values, connection, metakey)
        return result.all()

    async def fetchall(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Sequence[sa.Row]:
        """Alias to .all()"""
        return await self.all(query, values, connection, metakey)

    async def first(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row|None:
        """Get one (first/top) record from query. Returns None if no records found"""
        result = await self.execute(query, values, connection, metakey)
        return result.first()

    async def fetchone(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row|None:
        """Alias to .first()"""
        result = await self.execute(query, values, connection, metakey)
        return result.fetchone()

    async def one(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row:
        """Get one record from query. Throws Exception if no data found or querying more than one record"""
        result = await self.execute(query, values, connection, metakey)
        return result.one()

    async def one_or_none(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row|None:
        """Get one record from query.  Returns None if no record found.  Throws Exception of querying more than one record"""
        result = await self.execute(query, values, connection, metakey)
        return result.one_or_none()

    async def scalars(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Sequence[Any]:
        """Get many scalar values from query.  Returns empty List if no records found. If selecting multiple columns, returns List of FIRST column only."""
        result = await self.execute(query, values, connection, metakey)
        return result.scalars().all()

    async def scalar(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Any|None:
        """Get a single scalar value from query. Returns None if no record found.  Returns first (top) if more than one record found"""
        result = await self.execute(query, values, connection, metakey)
        return result.scalar()

    async def scalar_one(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Any:
        """Get a single scalar value from query.  Throws Exception if no data found or if querying more than one record"""
        result = await self.execute(query, values, connection, metakey)
        return result.scalar_one()

    async def scalar_one_or_none(self,
        query: sa.Select|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> Any|None:
        """Get a single scalar value from query.  Returns None if no record found.  Throws Exception if querying more than one record"""
        result = await self.execute(query, values, connection, metakey)
        return result.scalar_one_or_none()

    async def insertmany(self,
        query: sa.Insert|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> List[sa.Row]:
        """Bulk insert many rows, returning bulk primary keys (for databases that support INSERT..RETURNING)"""
        # For bulk returns see https://docs.sqlalchemy.org/en/20/core/connections.html#engine-insertmanyvalues
        result = await self.execute(query, values, connection, metakey)
        return result.inserted_primary_key_rows

    async def insertone(self,
        query: sa.Insert|str,
        values: Optional[Sequence[Mapping[str, Any]] | Mapping[str, Any]] = None,
        connection: Optional[str] = None,
        metakey: Optional[str] = None
    ) -> sa.Row:
        """Insert one row, returning the one rows PK (as a tuple in case of dual PKs)"""
        result = await self.execute(query, values, connection, metakey)
        return result.inserted_primary_key


# IoC Class Instance
# Not to be imported by the public from here.
# Use the uvicore.db singleton global instead.



#from __future__ import annotations
from abc import ABC, abstractmethod, abstractclassmethod
from typing import Optional, List, Generic, TypeVar
from uvicore.contracts import OrmQueryBuilder

E = TypeVar("E")

class AutoApi(Generic[E], ABC):

    @abstractmethod
    def orm_query(self) -> OrmQueryBuilder[OrmQueryBuilder, E]:
        """Start a new Uvicore ORM Model QueryBuilder Query"""



import re
import math
import uvicore
from os import stat
from pydantic import BaseModel, Field
from datetime import datetime, timezone
from uvicore.support.dumper import dump
from uvicore.support.module import load
from uvicore.http.request import Request
from pydantic.generics import GenericModel
from starlette.templating import _TemplateResponse
from prettyprinter import pretty_call, register_pretty
from starlette.background import BackgroundTask as _BackgroundTask
from uvicore.typing import Optional, Any, Dict, List, Generic, TypeVar

# Proxy starlette and fastapi response APIs
# Usage: from uvicore.http.response import FileResponse
#    Or: from uvicore.http import response
#        response.File
# See https://fastapi.tiangolo.com/reference/responses/
from starlette.responses import Response
from starlette.responses import FileResponse, FileResponse as File
from starlette.responses import HTMLResponse, HTMLResponse as HTML
from starlette.responses import JSONResponse, JSONResponse as JSON
from fastapi.responses import UJSONResponse, UJSONResponse as UJSON
from fastapi.responses import ORJSONResponse, ORJSONResponse as ORJSON
from starlette.responses import PlainTextResponse, PlainTextResponse as Text
from starlette.responses import RedirectResponse, RedirectResponse as Redirect
from starlette.responses import StreamingResponse, StreamingResponse as Stream


# Entity Generic
# Using Pydantic GenericModel for Generic OpenAPI schemas
# See https://medium.com/@jkishan421/building-dynamic-api-responses-with-generics-in-fastapi-972fa1f52d54 for details
E = TypeVar("E")


# Get our current template system from the IoC
#templates = uvicore.ioc.make('uvicore.http.templating.jinja.Jinja')
#templates = uvicore.ioc.make('uvicore.http.templating.engine.Templates')
templates = uvicore.ioc.make('uvicore.templating.engine.Templates')
#templates = uvicore.ioc.make('Templates') # Fixme when you impliment other templating engines, if ever


# Cached composer->view matches, a slight performance optimization found by wrk benchmarks
# This is because re.search and load() is expensive, no need to do it over and over.  Just
# do it once for each unique view and cache the found composer modules.
cached_composers: Dict[str, List] = {}


def utc_now():
    """UTC date factory for pydantic Field default_factory on RequestDate"""
    return datetime.now(timezone.utc)


@uvicore.service()
async def View(
    name: str,
    context: dict = {},
    status_code: int = 200,
    headers: dict = None,
    media_type: str = None,
    background: _BackgroundTask = None,
) -> _TemplateResponse:

    # Pull request out of context (which is always present as it is required for response.View())
    request: Request = context.get('request');

    # Convert context into SuperDict so we can merge in view composer context
    context = Dict(context)

    # Get all view composer modules
    composer_modules = []
    if name in cached_composers:
        # Cached composer(s) for this view were found, use the cached module
        #dump('Found cached composer for ' + name)
        composer_modules = cached_composers[name]
    else:
        # No cached composer yet found for this view.  Loop all view composers and re.search
        # the wildcards and dynamically load() the found composers.  If no composer found
        # still set the cache, but set to [] so we never attempt a re.search again
        found = False
        view_name = name.split('.')[0]
        for (composer_module, composer_views) in uvicore.config.uvicore.http.view_composers.items():
            for composer_view in composer_views:
                if (composer_view == '*'): composer_view = '.*'
                if re.search(composer_view, view_name):
                    #dump('Found UNcached composer for ' + name)
                    # Found a view composer matching this view name wildcard
                    found = True

                    # Ensure empty list
                    if name not in cached_composers: cached_composers[name] = []
                    try:
                        # Load the matched composer module, cache the module, but do not instantiate it
                        composer = load(composer_module).object
                        cached_composers[name].append(composer)
                        composer_modules.append(composer)
                    except:
                        # Composer module not found, bad config or missing file, silently ignore
                        pass
                    # Don't 'break' on match because there can be multiple composers matched to a view
        if not found:
            # No matching composer, still set cached_composers so we never try this again as it will NEVER match
            #dump('No composer cound, adding blahk [] to cache for ' + name)
            cached_composers[name] = []

    # Load all composers that match this view and merge in context
    for composer_module in composer_modules:
        # Load composer and merge the return using .defaults() to ensure
        # the view wins in the override battle over the composer
        composer = composer_module(request, name, context, status_code, headers, media_type)
        context.defaults(await composer.compose())

    #dump(name, cached_composers, composer_modules)

    # Render the template
    return templates.render_web_response(
        name=name,
        context=context,
        status_code=status_code,
        headers=headers,
        media_type=media_type,
        background=background
    )


@uvicore.service()
class APIResponse(GenericModel, Generic[E]):
    """Uvicore API Response"""
    # Careful, the text in """ above shows up in OpenAPI docs!
    api_version: Optional[str] = "1"
    request_date: Optional[datetime] = Field(default_factory=utc_now)
    response_date: Optional[datetime]
    response_ms: Optional[int]
    paged_response: bool = False
    result_count: Optional[int] = 0
    total_count: Optional[int] = 0
    page_num: Optional[int] = 0
    total_pages: Optional[int] = 0
    data: Optional[E]

    @staticmethod
    def begin() -> 'APIResponse':
        return APIResponse()

    @staticmethod
    def start() -> 'APIResponse':
        return APIResponse()

    @staticmethod
    def create() -> 'APIResponse':
        return APIResponse()


    def render(self, data: E, *, total_count: Optional[int] = None, page: int = 0, page_size: int = 0):
        """Render the API Response"""
        self.api_version = uvicore.config('app.api.version') or "1.0"
        self.response_date = utc_now()
        self.response_ms = int((self.response_date - self.request_date).total_seconds() * 1000)
        # If total_count is passed, we are using a paged response
        if total_count is not None:
            self.paged_response = True
            self.total_count = total_count
            self.page_num = page
            self.total_pages = math.ceil(total_count / page_size) if page_size > 0 else 0
        if isinstance(data, list): self.result_count = len(data)
        self.data = data
        return self

    def build(self, data: E, *, total_count: Optional[int] = None, page: int = 0, page_size: int = 0):
        """Alias to render()"""
        return self.render(data=data, total_count=total_count, page=page, page_size=page_size)

    def send(self, data: E, *, total_count: Optional[int] = None, page: int = 0, page_size: int = 0):
        """Alias to render()"""
        return self.render(data=data, total_count=total_count, page=page, page_size=page_size)

    def __call__(self, data: E, *, total_count: Optional[int] = None, page: int = 0, page_size: int = 0):
        """Alias to render()"""
        return self.render(data=data, total_count=total_count, page=page, page_size=page_size)


@uvicore.service()
class APIErrorResponse(BaseModel):
    """Uvicore API Error Response"""
    # Careful, the text in """ above shows up in OpenAPI docs!
    # Remember exception is REMOVED if not running in app.debug=True mode!
    status_code: Optional[int]
    message: Optional[str]
    detail: Optional[str]
    exception: Optional[str]
    extra: Optional[Any]


@register_pretty(APIResponse)
def pretty_entity(value, ctx):
    return pretty_call(ctx, APIResponse, **value.__dict__)

# @register_pretty(APIResult)
# def pretty_entity(value, ctx):
#     return pretty_call(ctx, APIResult, **value.__dict__)

@register_pretty(APIErrorResponse)
def pretty_entity(value, ctx):
    return pretty_call(ctx, APIErrorResponse, **value.__dict__)



import uvicore
import inspect
import fnmatch
from uvicore.typing import Optional, List, Tuple, Dict, Any, OrderedDict, Union
from uvicore.http.routing.api_router import ApiRouter
from uvicore.http.routing.router import Routes as Controller
from uvicore.http.request import Request
from uvicore.http.response import JSON
from uvicore.support import str as string
from uvicore.support.dumper import dump, dd
from uvicore.support import module
from uvicore.http.exceptions import HTTPException, NotFound
from uvicore.contracts import UserInfo
from uvicore.http.routing import Guard
from uvicore.orm import Model as OrmModel
from uvicore.http.routing.auto_api import AutoApi
from uvicore.support.collection import setvalue
from pydantic import BaseModel as PydanticModel
from pydantic import BaseModel

from starlette.responses import JSONResponse

# -where
# -or_where
# -filter
# -or_filter
# -order_by
# -sort
# -page and page_size - in vue and uvicore orm these are actually done with limit() and offset()
# Add a find for named finds still returning one


# key_by
# cache
# show_writeonly


class HTTPMessage(BaseModel):
  #status_code: int
  message: str
  #detail: Optional[str]
  #extra: Optional[Dict]


class DeleteQuery(PydanticModel):
    where: Optional[dict[str, Union[str, List]]]

    class Config:
        schema_extra = {
            "example": {
                "where": {
                    "creator_id": 1,
                    "title": ["in", ["title1", "title2"]],
                    "body": ["like", "%delete me%"],
                },
            }
        }

@uvicore.service()
class ModelRoutes:
    """Dynamic Model CRUD Routes"""

    @property
    def log(self):
        return uvicore.log.name('uvicore.model_router')

    def register(self, Model: OrmModel, route: ApiRouter, path: str, tags: List, scopes: List):
        """Build dynamic model CRUD routes"""

        #@route.get('/' + path, inherits=AutoApi.getsig, response_model=List[Model], tags=tags, scopes=[Model.tablename + '.read'] if scopes is None else scopes)
        @route.get(
            path='/' + path,
            inherits=AutoApi.getsig,
            response_model=Union[List[Model], Model, Dict],
            tags=tags,
            scopes=scopes['read'],
            summary="List multiple {}".format(Model.tablename),
            description="List one or more {} ({}) using the optional filtering parameters.".format(Model.tablename, Model.modelfqn),
        )
        async def get_all(**kwargs):
            api = AutoApi(Model, scopes, **kwargs).guard_relations()
            query = api.orm_query()

            # Run ORM query for results
            try:
                if api.find:
                    results = await query.find(**api.find)
                else:
                    results = await query.get()
                return results
            except Exception as e:
                raise HTTPException(
                    status_code=500,
                    detail="Error in query builder, most likely an unknown column or query parameter.",
                    exception=str(e)
                )

        @route.get(
            path='/' + path + '/{id}',
            inherits=AutoApi.findsig,
            # responses={
            #     404: {
            #         'model': HTTPMessage,
            #         #'description': 'xyz here',
            #         #'content': {"application/json": HTTPMessage},
            #     },
            # },
            response_model=Model,
            tags=tags,
            scopes=scopes['read'],
            summary="Get one {} by primary key".format(Model.tablename),
            description="Get a single {} ({}) by primary key.".format(Model.tablename, Model.modelfqn),
        )
        async def find_one(id: Union[str,int], **kwargs):
            api = AutoApi(Model, scopes, **kwargs).guard_relations()
            query = api.orm_query()

            # Run ORM query for results
            try:
                results = await query.find(id)
            except:
                raise HTTPException(500, str("Error in query builder, most likely an unknown column or query parameter."))

            if results:
                return results
            else:
                #raise NotFound()
                return JSONResponse(status_code=404, content={"message": "Item not found"})


        @route.post(
            path='/' + path,
            response_model=Union[Model, List[Model]],
            tags=tags,
            scopes=scopes['create'],
            summary="Create new {}".format(Model.tablename),
            description="Create one or more {} ({}) without nested relations by POSTING a valid and complete Model.".format(Model.tablename, Model.modelfqn),
        )
        #async def post(request: Request, item: Model):
        async def create(request: Request, items: Union[Model, List[Model]]):
            # NOTES
            # How do I check each child relations permissions, there is no includes
            # Would have to flip each item key and check if its a relation?
            # Then check if user has access to that relation?

            # Insert into storage and return primary key inserted
            try:
                # Ensure list
                is_single = False
                if type(items) != list:
                    items = [items]
                    is_single = True

                for item in items:
                    pk = await Model.insert(item)

                    # If primary key is read_only, assume its an auto-incrementing pk
                    # If not read_only, its a manual pk like 'key'.
                    # Only set new pk result if pk is read_only.  Why? Because when pk is 'key'
                    # a string, encode/databases does not return the new pk, just returns 1 every time.
                    if Model.modelfield(Model.pk).read_only == True:
                        setvalue(item, Model.pk, pk)

                # Return inserted item
                if is_single:
                    return items[0]
                else:
                    return items

            except Exception as e:
                raise HTTPException(500, str(e))


        @route.post(
            path='/' + path + '/with_relations',
            response_model=Union[Dict, List[Dict]],
            tags=tags,
            scopes=scopes['create'],
            summary="Create new {} with nested relations".format(Model.tablename),
            description="Create one or more {} ({}) with deeply nested relations by POSTING an object/array.  Uvicore must disable model validation when passing a complex nested relational object.  It is therefore up to you to ensure an accurate object is passed.".format(Model.tablename, Model.modelfqn),
        )
        async def create_with_relations(request: Request, items: Union[Dict, List[Dict]]):
            # NOTES
            # How do I check each child relations permissions, there is no includes
            # Would have to flip each item key and check if its a relation?
            # Then check if user has access to that relation?

            # Insert into storage and return primary key inserted
            try:
                # Ensure list
                is_single = False
                if type(items) != list:
                    items = [items]
                    is_single = True

                for item in items:
                    pk = await Model.insert_with_relations(item.copy())

                    # If primary key is read_only, assume its an auto-incrementing pk
                    # If not read_only, its a manual pk like 'key'.
                    # Only set new pk result if pk is read_only.  Why? Because when pk is 'key'
                    # a string, encode/databases does not return the new pk, just returns 1 every time.
                    if Model.modelfield(Model.pk).read_only == True:
                        setvalue(item, Model.pk, pk)

                # Return inserted items
                #dump(item)
                if is_single:
                    return items[0]
                else:
                    return items

            except Exception as e:
                raise HTTPException(500, str(e))


        @route.put(
            path='/' + path + '/{id}',
            response_model=Model,
            tags=tags,
            scopes=scopes['update'],
            summary="Update one complete {} by primary key".format(Model.tablename),
            description="Update a single {} ({}) by primary key by PUTTING a valid and complete Model.".format(Model.tablename, Model.modelfqn),
        )
        async def update_one(request: Request, id: Union[str,int], item: Model):
            # PUT is used to UPDATE an EXISTNIG record.
            # The PUT object must be a complete object. This is why 'item' is a Model, not a Dict
            try:
                result = await Model.query().find(id)
            except Exception as e:
                raise HTTPException(500, str(e))

            if result:
                # PUT requires a complete item.  It is not partial, it is not merged like a PATCH
                # So we take the entire "item" and simply use the PK from results to update the right record
                setattr(item, Model.pk, getattr(result, Model.pk))
                try:
                    await item.save()
                    return item
                except Exception as e:
                    raise HTTPException(500, str(e))
            else:
                raise NotFound('{} {} not found'.format(Model.modelname, id))


        @route.patch(
            path='/' + path + '/{id}',
            response_model=Model,
            tags=tags,
            scopes=scopes['update'],
            summary="Update one partial {} by primary key".format(Model.tablename),
            description="Update a single {} ({}) by primary key by PATCHING a partial object.  All fields in the object are optional.".format(Model.tablename, Model.modelfqn),
        )
        async def update_one_partial(request: Request, id: Union[str,int], item: Dict):
            # PATCH is used to UPDATE an EXISTNIG record.
            # The PATCH may be a partial object as it is merged with the existing object before being saved.
            # This is why 'item' must be a Dict, not a Model as pydantic would complain about missing fields.
            # All fields in a PATCH are optional.
            try:
                result = await Model.query().find(id)
            except Exception as e:
                raise HTTPException(500, str(e))

            if result:
                # PUT requires a complete item.  It is not partial, it is not merged like a PATCH
                # So we take the entire "item" and simply use the PK from results to update the right record
                for (key, value) in item.items():
                    setattr(result, key, value)
                try:
                    await result.save()
                    return result
                except Exception as e:
                    raise HTTPException(500, str(e))
            else:
                raise NotFound('{} {} not found'.format(Model.modelname, id))




        @route.delete(
            path='/' + path + '/{id}',
            # responses={
            #     404: {
            #         'model': HTTPMessage,
            #         'description': 'xyz here'
            #     },
            # },
            response_model=Model,
            tags=tags,
            scopes=scopes['delete'],
            summary="Delete one {} by primary key".format(Model.tablename),
            description="Delete a single {} ({}) by primary key.".format(Model.tablename, Model.modelfqn),
        )
        async def delete_one(request: Request, id: Union[str,int]):
            # DELETE must act on a single resource ONLY /{id}.  It does NOT take a body/paylad at all.
            # If you want to BULK delete using a body/payload with a JSON query, use POST instead with
            # a new custom endpoint like POST /users/delete payload={"where": ...}
            try:
                result = await Model.query().find(id)
            except Exception as e:
                raise HTTPException(500, str(e))

            if result:
                #await result.unlink('tags')
                await result.delete()
                return result
            else:
                raise NotFound('{} {} not found'.format(Model.modelname, id))
                #return JSON(status_code=404, content={"message": "asdf"})

        # NO, not RESTful.  DELETE does NOT take a body/payload
        # Instead, if you want bulk deletes, create a POST /users/delete with a payload query
        # @route.delete(
        #     path='/' + path,
        #     response_model=Model,
        #     tags=tags,
        #     scopes=scopes['delete'],
        #     summary="Delete {} using a query".format(Model.tablename),
        #     description="Delete one or more {} ({}) using a query.".format(Model.tablename, Model.modelfqn),
        # )
        # async def delete_many(request: Request, query: DeleteQuery):
        #     # PATCH is used to UPDATE an EXISTNIG record.
        #     # The PATCH may be a partial object as it is merged with the existing object before being saved.
        #     # This is why 'item' must be a Dict, not a Model as pydantic would complain about missing fields.
        #     # All fields in a PATCH are optional.
        #     try:
        #         dump(query, query.where)
        #         api = AutoApi(Model, scopes, request=request, where=query.where)
        #         q = api.orm_query()
        #         dump(q.query)

        #         results = await q.delete()
        #         dump(results)
        #         #return results


        #         #result = await Model.query().find(id)
        #         #await result.unlink('tags')
        #         #await result.delete()
        #     except Exception as e:
        #         raise HTTPException(500, str(e))



@uvicore.controller()
class ModelRouter(Controller):
    """Automatic CRUD Model Router"""

    # Notice.  This is actually must a regular uvicore route controller!
    # Just like any of your other controllers.  The only difference
    # is the register() is dynamically registering CRUD endpoints
    # for a dynamic list of all your models!

    def __init__(self, package, scopes: List = None, include: List = None, exclude: List = None):
        """Init override to add custom parameter"""

        # To add options from your routes/api.py use
        # route.include(ModelRouter, options={'scopes': []})
        super().__init__(package)
        self.scopes = scopes
        self.include = include
        self.exclude = exclude

    def register(self, route: ApiRouter):
        """Register API Route Endpoints"""

        # Get all models in tablename sorted order from Ioc bindings
        # This obeys self.include and self.exclude
        models = self._get_all_ioc_models()

        # Get source code for ModelRouter
        #modelroute = inspect.getsource(ModelRoute)
        #modelroute += "ModelRoute().routes(Model, route, path, tags, scopes)"

        # Loop each sorted model and dynamically add a ModelRouter
        for key, binding in models.items():

            # Get model information from Ioc binding
            Model = binding.object
            tablename = Model.tablename
            path = tablename
            tags = [string.ucbreakup(path)]
            permissions = Model.tablename

            # Get CRUD scopes plus any custom scopes
            scopes = self._get_scopes(tablename)
            #dump("FULL SCOPES: ", scopes)

            # Instantiate our ModelRouter class with this one Uvicore model
            ModelRoutes().register(Model, route, path, tags, scopes)


            # OLD - This was when I was using FastAPIs route directly instead
            # of my new uvicore router abstraction.  Now with uvicore router
            # all of this dynamic inspect exec() is not needed!!!
            # ------------------------------------------------------------------
            # Dynamically instantiate ModelRouter from source code and exec
            # passing in the proper globals.  Why?  Why not just instantiate
            # the class and be done?  Becuase pydantic doesn't understand
            # type hinting that way.  For example post(entity: Model) does NOT
            # work as the type hinter errors on the Model.  I tried every other way
            # and type hinting just does not work as expected.  Even with Generics[E]
            # Not sure why exactly. This dynamic execution however works for now.
            # One caveat to this is you cannot override and EXTEND the ModelRoute
            # class.  You can override, but you must re-impliment the entire class
            # and not just extend it.
            # exec(modelroute, {
            #     'Model': Model,
            #     'route': route,
            #     'path': path,
            #     'tags': tags,
            #     'uvicore': uvicore,
            #     'typing': typing,
            #     'UserInfo': UserInfo,
            #     'Guard': Guard,
            #     'Request': Request,
            #     'HTTPException': HTTPException,
            #     'PermissionDenied': PermissionDenied,
            #     'scopes': self.scopes,
            #     'dump': dump,
            #     'json': json,
            #     'params': params,
            # })

        # Return router
        return route

    def _get_all_ioc_models(self) -> OrderedDict:
        """Get all models in tablename sorted order from Ioc bindings"""
        models = OrderedDict()
        unsorted_models = {}
        model_bindings = uvicore.ioc.binding(type='model', include_overrides=False)

        all_model_keys = set(model_bindings.keys())
        model_keys = []

        # If include, only include these models (wildcards accepted)
        if self.include:
            for include in self.include:
                model_keys.extend(fnmatch.filter(all_model_keys, include))

            # Ensure unique List in case of double include entries matching double
            model_keys = list(set(model_keys))
        else:
            # No includes, use all models
            model_keys = all_model_keys

        # If exclude, remove them from the list (wildcards accepted)
        if self.exclude:
            for exclude in self.exclude:
                model_keys = [x for x in model_keys if x not in fnmatch.filter(model_keys, exclude)]

        # Now we have our exact models as list of python module paths from the IOC
        # Now we get actual IOC bindings except our Dict Key is the TABLENAME, not module path
        for model in model_keys:
            tablename = model_bindings[model].object.tablename
            if tablename not in models:
                unsorted_models[tablename] = model_bindings[model]

        # Now we can sort our unsorted_model keys (which are tablenames)
        tablenames = sorted(unsorted_models.keys())

        # Add all sorted models to our final models OrderedDict
        for tablename in tablenames:
            models[tablename] = unsorted_models[tablename]

        return models

    def _get_scopes(self, tablename) -> Dict:
        """Get scopes from options override or default automatic model CRUD scopes"""

        # If custom scopes are passed, add those to every CRUD scope
        # Custom scopes are defined in self.scopes as either a List or Dict
        # If list, apply entire list to all CRUD.  If Dict, only apply to proper CRUD key
        # If scopes is None, no custom scopes are used, so we apply our default model CRUD scopes

        scopes = {}
        scopes['create'] = []
        scopes['read'] = []
        scopes['update'] = []
        scopes['delete'] = []
        scopes['overridden'] = False

        # Note: If you want to ADD other scopes to the automatic model CRUD scopes
        # use a @route.group(scopes=['additional']) instead of the router
        # .include() or .controller() options dict.

        if self.scopes is None:
            # No custom scopes defined, add automatic model CRUD scopes
            scopes['create'].append(tablename + '.create')
            scopes['read'].append(tablename + '.read')
            scopes['update'].append(tablename + '.update')
            scopes['delete'].append(tablename + '.delete')
        else:
            # Custom scopes defined, use these scopes insetead of automatic model CRUD scopes
            scopes['overridden'] = True
            if type(self.scopes) == list:
                # Custom scopes is a list, so apply entire list to each CRUD section
                # Must copy() each list or all will be references to same list
                scopes['create'] = self.scopes.copy()
                scopes['read'] = self.scopes.copy()
                scopes['update'] = self.scopes.copy()
                scopes['delete'] = self.scopes.copy()
            else:
                # Must copy() each list or all will be references to same list
                scopes['create'] = self.scopes.copy().get('create') or []
                scopes['read'] = self.scopes.copy().get('read') or []
                scopes['update'] = self.scopes.copy().get('update') or []
                scopes['delete'] = self.scopes.copy().get('delete') or []
        return scopes



import uvicore
from uvicore.auth.support import password
from uvicore.support.dumper import dump, dd
from uvicore.auth.models.permission import Permission

@uvicore.seeder()
async def seed():
    #from uvicore.auth.models

    # NOTICE, each model role is seeded from the db seeder command
    # This is where I add additional permissions

    await Permission.insert([
        {
            'entity': None,
            'name': 'admin',
        },
    ])

    # NO, we do this from db seeder command now

    # uvicore.log.item('Seeding table permissions')

    # entities = [
    #     uvicore.db.tablename('auth.group_roles'),
    #     uvicore.db.tablename('auth.groups'),
    #     uvicore.db.tablename('auth.permissions'),
    #     uvicore.db.tablename('auth.role_permissions'),
    #     uvicore.db.tablename('auth.roles'),
    #     uvicore.db.tablename('auth.user_groups'),
    #     uvicore.db.tablename('auth.user_roles'),
    #     uvicore.db.tablename('auth.users'),
    # ]

    # permissions = [
    #     'create',   # Django add
    #     'read',     # Django view
    #     'update',   # Django change
    #     'delete',   # Django delete
    # ]

    # bulk = []
    # for entity in entities:
    #     for permission in permissions:
    #         bulk.append(Permission(entity=entity, name=entity + '.' + permission))

    # await Permission.insert(bulk)



# type: ignore
from . import contracts
from uvicore.typing import Dict, TYPE_CHECKING
from uvicore.foundation.decorators import event, job, model, seeder, service, table, provider, routes, controller, composer


# Uvicore version.  Also available in app.version
__version__ = '0.3.13'

# Foundational
ioc: contracts.Ioc = None
events: contracts.Dispatcher = None
jobs: contracts.JobDispatcher = None
app: contracts.Application = None
config: contracts.Config = None
log: contracts.Logger = None
cache: contracts.Cache = None

# Optional
if TYPE_CHECKING:
    db: contracts.Database = None
else:
    db = None


def bootstrap(app_config: Dict, path: str, is_console: bool) -> None:
    """A pre-bootstrap method to setup and call the main uvicore application bootstrap"""

    # Order of imports is critical for proper IoC binding and override
    import uvicore
    from uvicore.support.dumper import dump, dd

    # Ensure app_config is a Uvicore Types Dict
    app_config = Dict(app_config)

    # Initialize the singleton IoC container
    # Before importing Application and Dispatcher which lets the IoC from app_config
    # swap even the earliest of core services
    from .container.ioc import Ioc
    uvicore.ioc = Ioc(app_config)

    # Import Application (which is an IoC singleton) and set uvicore.app global
    from uvicore.foundation.application import Application
    #uvicore.app = uvicore.ioc.make('uvicore.foundation.application.Application') # Also works
    uvicore.app = Application

    # Import Event Dispatcher (which is an IoC singleton) and set uvicore.events global
    # This cannot be a service provider because events are fired BEFORE any service providers
    # are ever loaded.
    from uvicore.events.dispatcher import Dispatcher
    uvicore.events = Dispatcher

    # Import Job Dispatcher
    from uvicore.jobs.dispatcher import Dispatcher as JobDispatcher
    uvicore.jobs = JobDispatcher

    # Why can 'config' be a service provider?  Because it's always the FIRST service provider.
    # So all other providers after that have access to register their own configs.

    # Bootstrap the actual uvicore Application
    uvicore.app.bootstrap(app_config, path, is_console)



import os
import uvicore
from uvicore import log
from uvicore.support import str
from uvicore.console import command, argument
from uvicore.support.dumper import dump, dd
from uvicore.support.schematic import Schematic


@command()
@argument('name')
async def command(name: str):
    """
    \b
    Generate a new CLI command schematic...
    \b
    USAGE:
        Commands should be lower_underscore and SINGULAR (plural is OK)
        Remember to manually add the command to your service provider!
    \b
        ./uvicore gen command welcome
        ./uvicore gen command process
        ./uvicore gen command scan_files
    """

    # Get calling package (main running app)
    package = uvicore.app.package(main=True)

    # Get stub (src)
    stub = os.path.dirname(__file__) + '/stubs/command.py'

    # Get destination for this filetype, considering the packages path customizations
    dest = package.folder_path('commands') + '/' + name + '.py'

    Schematic(
        type='command',
        stub=stub,
        dest=dest,
        replace = [
            ('xx_name', name)
        ]
    ).generate()

    # Get running package
    package = uvicore.app.package(main=True)

    log.nl()
    log.header('Add this to your Service Provider boot() or define_commands self.commands()')
    print("'{}': '{}.commands.{}.cli',".format(str.kebab(name), package.name, name))

    log.nl()
    log.notice('IF you do NOT have a self.commands() already in your Service Provider, add this')
    print("""self.register_cli_commands(
    group='{}',
    help='{} Commands',
    commands={{
        '{}': '{}.commands.{}.cli',
    }}
)""".format(
        package.short_name,
        str.studly(package.short_name),
        str.kebab(name),
        package.name,
        name,
    ))



import uvicore
from uvicore.typing import Any, Dict, Optional

# Smart exception should work regardless
has_http = True
try:
    from uvicore.http.exceptions import HTTPException as BaseException
    #from starlette.exceptions import HTTPException as BaseException
except ImportError as e:  # pragma: nocover
    BaseException = Exception
    has_http = False


class SmartException(BaseException):
    # Notice detail and status_code are reversed from my uvicore.http.exceptions.HTTPException
    # This is intentional as SmartExceptions are used as a mix of CLI and HTTP
    def __init__(self,
        detail: Optional[str],
        status_code: int = None,
        *,
        message: Optional[str] = None,
        exception: Optional[str] = None,
        extra: Optional[Dict] = None,
        headers: Optional[Dict[str, Any]] = None
    ) -> None:
        # Uvicore HTTP module is installed and we are running as HTTP
        if has_http and uvicore.app.is_http:
            # Call uvicores base HTTPException
            if not status_code: status_code = 500
            super().__init__(
                status_code=status_code,
                message=message,  # None message will show defailt HTTP status code phrase
                detail=detail,
                exception=exception,
                extra=extra,
                headers=headers,
            )
        else:
            # Uvicore HTTP module is not installed and/or we are not running as HTTP (from CLI)
            # Status code for CLI is still applicable, can use as bash exit code or general error code instead
            if not status_code: status_code = 1  # Bash catch all for general errors
            if not message: message = 'An error has occured'
            self.status_code = status_code
            self.message = message
            self.detail = detail
            self.exception = exception if uvicore.config.app.debug else None  # Hidden unless in debug mode
            self.extra = extra
            # No need for self.headers in CLI




import uvicore
from uvicore.support.dumper import dump, dd
from uvicore.exceptions import SmartException
from uvicore.console import command, argument, option


# ------------------------------------------------------------------------------
# Uvicore Command Schematic
# This schematic is filled with examples, a suppliment to the docs.
# Pick the best example for your use case and modify as needed!
# ------------------------------------------------------------------------------


# --------------------------------------------------------------------------
# Example: Basic command with on arguments or options
# --------------------------------------------------------------------------
@command()
async def cli():
    """xx_AppName xx_name CLI command"""
    try:
        print('Welcome to your new xx_AppName xx_name CLI command!')
    except SmartException as e:
        # Python exit() with any value means "error" in bash exit code speak!
        exit(e.detail)




# --------------------------------------------------------------------------
# Example: Command with arguments and options
# --------------------------------------------------------------------------
# @command(help="This is another place to set command help messages")
# @argument('id_or_name')
# @option('--tenant', help='Tenant')
# @option('--coin', default='BTC', help='Coin with Default')
# @option('--json', is_flag=True, help='Output results as JSON')
# async def get(id_or_name: str, tenant: str, coin: str, json: bool):
#     """This shows up as the commands help message"""
#     # ex: ./uvicore xx_name namearg --tenant bob --json
#     from uvicore.exceptions import SmartException
#     try:
#         # Do stuff
#         pass
#     except SmartException as e:
#         # Python exit() with any value means "error" in bash exit code speak!
#         exit(e.detail)



import os
import uvicore
from uvicore import log
from uvicore.support import str
from uvicore.support.dumper import dump, dd
from uvicore.support.schematic import Schematic
from uvicore.console import command, argument, option


@command()
@argument('name')
async def controller(name: str):
    """
    \b
    Generate a new HTTP Web controller schematic...
    \b
    USAGE:
        Web Controllers should be lower_understore and SINGULAR
    \b
        ./uvicore gen controller home
        ./uvicore gen controller about
        ./uvicore gen controller contact_us
    """
    # Get calling package (main running app)
    package = uvicore.app.package(main=True)

    # Get stub (src)
    stub = os.path.dirname(__file__) + '/stubs/controller.py'

    # Get destination for this filetype, considering the packages path customizations
    dest = package.folder_path('controllers') + '/' + name + '.py'

    Schematic(
        type='controller',
        stub=stub,
        dest=dest,
        replace = [
            ('xx_controllername', name),
            ('xx_ControllerName', str.studly(name)),
        ]
    ).generate()


@command()
@argument('name')
async def api_controller(name: str):
    """
    \b
    Generate a new HTTP API controller schematic...
    \b
    USAGE:
        API Controllers should be lower_understore and SINGULAR
    \b
        ./uvicore gen controller user
        ./uvicore gen controller user_detail
        ./uvicore gen controller post
        ./uvicore gen controller post_tag
    """

    # Get calling package (main running app)
    package = uvicore.app.package(main=True)

    # Get stub (src)
    stub = os.path.dirname(__file__) + '/stubs/api_controller.py'

    # Get destination for this filetype, considering the packages path customizations
    dest = package.folder_path('api') + '/' + name + '.py'

    Schematic(
        type='api_controller',
        stub=stub,
        dest=dest,
        replace = [
            ('xx_controllername', name),
            ('xx_ControllerName', str.studly(name)),
        ]
    ).generate()


@command()
@argument('name')
async def composer(name: str):
    """
    \b
    Generate a new web view composer schematic...
    \b
    USAGE:
        View Composers should be lower_understore and SINGULAR
    \b
        ./uvicore gen composer layout
        ./uvicore gen composer side_nav
    """

    # Get calling package (main running app)
    package = uvicore.app.package(main=True)

    # Get stub (src)
    stub = os.path.dirname(__file__) + '/stubs/composer.py'

    # Get destination for this filetype, considering the packages path customizations
    dest = package.folder_path('view_composers') + '/' + name + '.py'

    Schematic(
        type='composer',
        stub=stub,
        dest=dest,
        replace = [
            ('xx_composername', name),
            ('xx_ComposerName', str.studly(name)),
        ]
    ).generate()

    # Get running package
    package = uvicore.app.package(main=True)

    log.nl()
    log.header('Add this to your Service Provider boot() or define_views()')
    print("self.composers('{}/*', '{}.http.composers.{}.{}')".format(package.short_name, package.name, name, str.studly(name)))



import pickle

import uvicore
from time import time
from uvicore.typing import Dict, Any, Callable, Union, List, Tuple
from uvicore.support.dumper import dump, dd
from uvicore.contracts import Cache as CacheInterface
from uvicore.cache.manager import Manager

@uvicore.service()
class Array(CacheInterface):
    def __init__(self, manager: Manager, store: Dict):
        self.manager = manager
        self.prefix = store.prefix
        self.seconds = store.seconds
        self.items = {}
        self.items_ttl = {}

    def connect(self, store: str = None) -> CacheInterface:
        """Connect to a cache backend store"""
        return self.manager.connect(store)

    def store(self, store: str = None) -> CacheInterface:
        """Alias to connect"""
        return self.connect(store)

    async def has(self, key: str) -> bool:
        """Check if key exists"""
        key = self._prepair(key)
        return key in self.items.keys()

    async def get(self, key: Union[str, List], *, default: Any = None) -> Any:
        """Get one or more key values if exists else return default value"""
        keys = self._prepair(key)
        if type(keys) == list:
            values = Dict()
            for key in keys:
                return_key = key[len(self.prefix):]
                if self._has(key):
                    # Item exists, get it
                    values[return_key] = self._deserialize(self.items[key])
                else:
                    # Item does not exist, set default
                    values[return_key] = default
            return values
        else:
            if self._has(keys):
                # Item exists, get it
                return self._deserialize(self.items[keys])
            else:
                # Item does not exist, set default
                return default

    async def remember(self, key: Union[str, Dict], callback: Union[Callable, Any] = None, *, seconds: int = None) -> Any:
        """Get a key if exists, if not SET the key to callback value"""
        keys = self._prepair(key)
        if type(key) != dict: keys = {keys:callback}
        value = {}
        for key, callback in keys.items():
            if self._has(key):
                # Item exists, simply return value
                value[key] = await self.get(key)
            else:
                # Item does not exist, set value based on callback return
                if callable(callback):
                    value[key] = await callback()
                else:
                    value[key] = callback
                await self.put(key, value[key], seconds=seconds)

        if type(key) != dict:
            # Single key, single return
            return value[key]

        # Dict key, dict return
        return value

    async def put(self, key: Union[str, Dict], value: Any = None, *, seconds: int = None) -> None:
        """Put one or more key/values in cache with optional expire in seconds (0=never expire)"""
        keys = self._prepair(key)
        if seconds is None: seconds = self.seconds
        if type(keys) != dict: keys = {keys:value}
        for (key, value) in keys.items():
            self.items[key] = self._serialize(value)
            if seconds == 0 and key in self.items_ttl:
                del self.items_ttl[key]
            elif seconds > 0:
                self.items_ttl[key] = self._now() + seconds

    async def pull(self, key: Union[str, Dict]) -> Any:
        """Get one or more key values from cache them remove them after"""
        keys = self._prepair(key)
        value = await self.get(key)
        await self.forget(keys)
        return value

    async def add(self, key: str, value: Any, *, seconds: int = None) -> bool:
        """Put a single value in cache only if not exists"""
        key = self._prepair(key)
        if self._has(key):
            # Item already exists, return False for NOT added
            return False
        else:
            # Item does not exist, set value and return True
            await self.put(key, value)
            return True

    async def touch(self, key: str, *, seconds: int = None) -> bool:
        """Touch a key, if seconds are provided, also reset expire TTL"""
        key = self._prepair(key)
        if seconds is None: seconds = self.seconds
        if self._has(key) and seconds is not None:
            self.items_ttl[key] = self._now() + seconds
            return True
        return False

    async def increment(self, key, by: int = 1, *, seconds: int = None) -> int:
        """Increment a key by integer specified.  If key not exists, sets key to increment value"""
        key = self._prepair(key)
        if seconds is None: seconds = self.seconds
        value = 0
        if self._has(key): value = await self.get(key)
        if type(value) == int:
            value += 1
            self.put(key, value, seconds=seconds)
        return value

    async def decrement(self, key, by: int = 1, *, seconds: int = None) -> int:
        """Decrement a key by integer specified.  If key not exists, sets key to decrement value"""
        key = self._prepair(key)
        if seconds is None: seconds = self.seconds
        value = 0
        if self._has(key): value = await self.get(key)
        if type(value) == int:
            value -= 1
            self.put(key, value, seconds=seconds)
        return value

    async def forget(self, key: Union[str, List]) -> None:
        """Delete a key from cache"""
        keys = self._prepair(key)
        if type(keys) != list: keys = [keys]
        for key in keys:
            if key in self.items:
                del self.items[key]
            if key in self.items_ttl:
                del self.items_ttl[key]

    async def flush(self) -> None:
        """Flush entire cache.  Only deletes keys with proper cache prefix."""
        delete = []
        for key in self.items.keys():
            if self.prefix in key:
                delete.append(key)
        for key in delete:
            del self.items[key]

    def _has(self, key: str) -> bool:
        # This is an internal _has() only.  Why?  Because the public facing
        # has is async def.  But it doesn't need to be.  Need to keep public
        # as async def to match interface as Redis and other backends must be
        # async def.

        # I don't use _prepair here because ._has() is often used
        # by many other internal method here.  Which means _prepair and therefore
        # _expire() will be called many times.  So instead I just deduce
        # the prefix right here.  This is only OK because _has() is INTERNAL
        # and all callers already check _expire through their _prepair call.
        # The public facing has() must use _prepair and therefore _expire or you
        # could say a key exists even if its technically expired.
        key = str(key)
        prefix = self.prefix
        if len(key) > len(prefix) and key[0:len(prefix)] == prefix: prefix = ''
        return (prefix + key) in self.items.keys()

    def _prepair(self, key: Union[str, List] = None) -> Union[str, List, Dict]:
        if key:
            if type(key) == list:
                # Check if prefix already added
                key0 = str(key[0])
                prefix = self.prefix
                if len(key0) > len(prefix) and key0[0:len(prefix)] == prefix: prefix = ''
                keys = [prefix + k for k in key]
                for k in keys:
                    self._expire(k)
                return keys
            elif type(key) == dict:
                key0 = [x for x in key.keys()][0]
                prefix = self.prefix
                if len(key0) > len(prefix) and key0[0:len(prefix)] == prefix: prefix = ''
                for k in key.keys():
                    self._expire(prefix + k)
                return {prefix + k:v for k,v in key.items()}
            else:
                # Check if prefix already added
                key = str(key)
                prefix = self.prefix
                if len(key) > len(prefix) and key[0:len(prefix)] == prefix: prefix = ''
                self._expire(prefix + key)
                return prefix + key
        else:
            return None

    def _expire(self, key):
        """Delete keys that are expired anytime they are accessed"""
        # If entry not in self.items_ttl, it never expires, keep it forever
        if key in self.items_ttl:
            if key in self.items.keys():
                now = self._now()
                if now >= self.items_ttl[key]:
                    del self.items[key]
                    del self.items_ttl[key]

    def _now(self) -> int:
         return int(time())

    def _serialize(self, value):
        return value
        #return pickle.dumps(value)

    def _deserialize(self, value):
        # Error here means value was never serialized.
        # Like with .increment and .decrement keys
        return value
        # try:
        #     return pickle.loads(value)
        # except:
        #     return value.decode()




import uvicore
from jwt import PyJWKClient, decode
from uvicore.support.hash import md5
from uvicore.contracts import UserInfo
from uvicore.support.dumper import dump, dd
from uvicore.http.request import HTTPConnection
from uvicore.typing import Dict, Optional, Union, Callable
from uvicore.auth.authenticators.base import Authenticator
from uvicore.http.exceptions import NotAuthenticated, InvalidCredentials, HTTPException


@uvicore.service()
class Jwt(Authenticator):
    """JWT bearer token authenticator"""

    # Notice:  Do not ever throw or return an error from authenticators.
    # If there is any auth problem (no headers, invalid or expired tokens, bad password)
    # then return True or False instead.  Any issues or invalid credentials causes the global
    # Authentication middleware to inject an Anonymous user, not to throw errors.
    # Any permissions errors happen at a route level when checking of the route
    # require an authenticated user or valid scope/role.

    # Return of False means this authentication method is not being attempted, try next authenticator
    # Return of True means this authentication method was being attempted, but failed validation, skip next authenticator
    # Return of User object means a valid user was found, skip next authenticator

    async def authenticate(self, request: HTTPConnection) -> Union[UserInfo, bool]:
        self.log.debug('JWT Authenticator')

        # Parse authorization header
        authorization, scheme, token = self.auth_header(request)
        if authorization: self.log.debug(authorization)

        # This authentication method not provided or attempted, goto next authenticator
        if not authorization or scheme != "bearer":
            # Return of False means this authentication method is not being attempted
            # goto next authenticator in stack
            self.log.debug('No Bearer found in Authorization header, goto next authenticator in stack')
            return False

        # Detect anonymous API access.
        # API gateways that allow anonymous access to an API will pass the request
        # downstream even if the token is missing, invalid or expired.
        # When they do, they add a header notifying downstream that auth has failed
        # but anonymous access is still allowed.  For Kong this header is x-anonymous-consumer
        # If the API gateway does not allow anonymous access, downstream won't even be proxied.
        #dump(request.headers)
        if (self.config.anonymous_header and self.config.anonymous_header in request.headers):
            # Anonymous header found, return True to denote Anonymous user and skip next authenticator
            self.log.debug('Anonymous header found, this is an Anonymous request')
            return True

        # Decode JWT with or without verification
        jwt = None

        # Verify JWT internally with uvicore
        if self.config.verify_signature:
            verify_signature_method = self.config.verify_signature_method.lower()
            self.log.debug(f"Validating JWT Signature using '{verify_signature_method}' method")

            # Verify JWT using JWKS
            if verify_signature_method == 'jwks':
                default_jwks_url = uvicore.config.app.auth.oauth2.base_url + uvicore.config.app.auth.oauth2.jwks_path
                found_consumer = False
                for (consumer_name, consumer) in self.config.consumers.items():
                    try:
                        # Cache decoded JWT to prevent hitting JWKS url often
                        async def query_jwks():
                            #dump('NO CACHE, TRY CONSUMER: ' + consumer_name)

                            # Get default or overridden jwks_url
                            jwks_url = consumer.jwks_url or default_jwks_url
                            self.log.debug('Verifying JWT via JWKS URL {}'.format(jwks_url))

                            # Get signing_key from jwks
                            jwks_client = PyJWKClient(jwks_url)
                            signing_key = jwks_client.get_signing_key_from_jwt(token)
                            return Dict(decode(
                                jwt=token,
                                key=signing_key.key,
                                audience=consumer.aud,
                                algorithms=consumer.algorithms
                            ))

                        # Get decoded JWT from cache or decode from JWKS URL signing_key
                        jwt = await uvicore.cache.remember(token, query_jwks, seconds=self.config.jwks_query_cache_ttl or 300)
                        found_consumer = True
                        self.log.debug('Found consumer via JWKS, name: {}, aud: {}'.format(consumer_name, consumer.aud))
                        break
                    except Exception as e:
                        # Means this one consumer loop JWK or secret not verified, OK, continue gracefully to next consumer
                        pass

                if not found_consumer:
                    self.log.debug('Could not validate JWT against any consumers vai JWKS.  Return True to denote Anonymous user and skip next authenticator')
                    return True

            # Verify JWT using pub key secrets
            else:
                for (consumer_name, consumer) in self.config.consumers.items():
                    try:
                        jwt = Dict(decode(
                            jwt=token,
                            key=consumer.secret,
                            audience=consumer.aud,
                            algorithms=consumer.algorithms
                        ))
                        self.log.debug('Found consumer via SECRET, name: {}, aud: {}'.format(consumer_name, consumer.aud))
                        break
                    except Exception as e:
                        self.log.debug('Issue validating JWT via SECRETS.  No valid secrets or audiences found.  Return True to denote Anonymous user and skip next authenticator')
                        self.log.debug(e)
                        return True

        # No JWT validation (should be pre-validated by external API Gateway like Kong)
        else:
            self.log.debug('External JWT verification via API Gateway is being used.  Uvicore will not validate JWT but may verify aud claims.')
            try:
                # Decode JWT without signature verification
                jwt = Dict(decode(token, options={"verify_signature": False}))

                # Validate aud claims if enabled
                if self.config.verify_audience:
                    # Get all allowed consumer audiences
                    audiences = [v['aud'] for (k,v) in self.config.consumers.items()]
                    if jwt.aud not in audiences:
                        # Audience mismatch, return True to denote Anonymous user and skip next authenticator
                        self.log.debug('No audience found in allowed consumers')
                        return True
            except Exception as e:
                self.log.debug('Issue decoding JWT without signature verification Ruturn True to denote Anonymous user and skip next authenticator')
                self.log.debug(e)
                return True

        # Dump JWT to DEBUG log
        self.log.debug('JWT: ' + str(jwt))

        # Map JWT based on our auth configs JWT authenticator auto_create_user_jwt_mapping
        mapped_jwt = {}
        for key, value in self.config.auto_create_user_jwt_mapping.items():
            if isinstance(value, Callable):
                mapped_jwt[key] = value(jwt)
            else:
                mapped_jwt[key] = value

        # Dump Mapped JWT to DEBUG log
        self.log.debug('Mapped JWT: ' + str(mapped_jwt))

        # Get user and validate credentials
        user: UserInfo = await self.retrieve_user(mapped_jwt['email'], None, self.config.provider, request, jwt=jwt)

        # User from valid JWT not found in uvicore OR not synced for first time (no uuid).
        # Auto create or update user if allowed in config
        if self.config.auto_create_user and (user is None or user.uuid is None):
            # User does not exist, create user
            if user is None:
                # Auto create new user in user provider
                await self.create_user(self.config.provider, request, **mapped_jwt)
            else:
                # User exists, but needs an initial sync
                await self.sync_user(self.config.provider, request, **mapped_jwt)

            # Re-pull user after creation
            user: UserInfo = await self.retrieve_user(mapped_jwt['email'], None, self.config.provider, request, jwt=jwt)

        # Periodically sync user and group info from JWT
        if (self.config.sync_user):
            # Example cache key mreschke.wiki::cache/users/mreschke@example.com/sync_user_ttl
            cache_key = "users/" + user.email + "/sync_user_ttl"
            cache_ttl = self.config.sync_user_ttl
            #if not await uvicore.cache.has(cache_key):
            await self.sync_user(self.config.provider, request, **mapped_jwt)
            await uvicore.cache.put(cache_key, 1, seconds=cache_ttl)

        # Return user.  If no user return True to denote Anonymous User and skip next authenticator
        return user or True



import uvicore
from uvicore.support.dumper import dump, dd
from dataclasses import dataclass
from uvicore.typing import Dict, Optional, Union, Callable
from uvicore.contracts.job import Job as JobInterface


@uvicore.service()
class Job(JobInterface):

    @classmethod
    @property
    def name(cls):
        """Get the name of this job"""
        name = str(cls).split("'")[1]
        return name

    @classmethod
    @property
    def description(cls):
        """Get the doc description of this job"""
        return cls.__doc__

    def dispatch(self):
        """Dispatch a Job Class"""
        return uvicore.jobs.dispatch(self)

    async def dispatch_async(self):
        """Dispatch an async Job Class"""
        return await uvicore.jobs.dispatch_async(self)

    async def codispatch(self):
        """Dispatch an async Job Class"""
        return await uvicore.jobs.dispatch_async(self)



import uvicore
from uvicore.package import Provider


@uvicore.provider()
class Mail(Provider):

    def register(self) -> None:
        """Register package into the uvicore framework.
        All packages are registered before the framework boots.  This is where
        you define your packages configs, IoC bindings and early event listeners.
        Configs are deep merged only after all packages are registered.  No real
        work should be performed here as it is very early in the bootstraping
        process and we have no clear view of the full configuration system."""

        pass


    def boot(self) -> None:
       """Bootstrap package into the uvicore framework.
        Boot takes place after ALL packages are registered.  This means all package
        configs are deep merged to provide a complete and accurate view of all
        configuration. This is where you register, connections, models,
        views, assets, routes, commands...  If you need to perform work after ALL
        packages have booted, use the event system and listen to the booted event:
        self.events.listen('uvicore.foundation.events.app.Booted', self.booted)"""

        pass



from __future__ import annotations

import uvicore
from typing import Dict, List, Optional
#from uvicore.auth.models.user import User
from uvicore.support.dumper import dd, dump
from uvicore.orm import Model, ModelMetaclass, Field
from xx_vendor.xx_appname.database.tables import xx_tablename as table
from uvicore.orm.fields import BelongsTo, BelongsToMany, HasMany, HasOne, MorphOne, MorphMany, MorphToMany


# ------------------------------------------------------------------------------
# Uvicore Database Model Schematic
# This schematic is filled with examples, a suppliment to the docs.
# Pick the best example for your use case and modify as needed!
# ------------------------------------------------------------------------------



# --------------------------------------------------------------------------
# Example: How to override the main uvicore.auth user table
# --------------------------------------------------------------------------
# AuthOverride = uvicore.ioc.make('uvicore.auth.models.user.User_BASE')
# @uvicore.model()
# class User(AuthOverride, Model['User'], metaclass=ModelMetaclass):
#     # your custom user fields
#     pass



@uvicore.model()
class xx_ModelName(Model['xx_ModelName'], metaclass=ModelMetaclass):
    """xx_AppName xx_ModelName Model"""

    # Database table definition
    # Optional as some models have no database table
    __tableclass__ = table.xx_TableName

    id: Optional[int] = Field('id',
        primary=True,
        description='xx_ModelName ID',
        #sortable=True,
        #searchable=True,
        read_only=True,
    )

    slug: str = Field('slug',
        description='URL Friendly xx_ModelName Title Slug',
    )

    title: str = Field('title',
        description='xx_ModelName Title',
    )

    # created_at: Optional[datetime] = Field('created_at',
    #     description='xx_ModelName Created at Datetime',
    #     read_only=True,
    # )

    # deleted: Optional[bool] = Field('deleted',
    #     description='xx_ModelName Deleted',
    #     default=False,
    # )



    # --------------------------------------------------------------------------
    # Example: One-To-One (Post has one info row)
    # --------------------------------------------------------------------------
    # info: Optional[PostInfo] = Field(None,
    #     description='Post Info Model',
    #     relation=HasOne('xx_vendor.xx_appname.models.post_info.PostInfo', foreign_key='post_id'),
    # )
        # PostInfo model would have the one-to-one inverse (info has one post)
        # post_id: int = Field('post_id',
        #     description="Post ID",
        # )
        # post: Optional[Post] = Field(None,
        #     description="Post Model",
        #     relation=BelongsTo('xx_vendor.xx_appname.models.post.Post')
        # )



    # --------------------------------------------------------------------------
    # Example:  One-To-Many Inverse (One Post has One Creator)
    # --------------------------------------------------------------------------
    # creator_id: int = Field('creator_id',
    #     description="Post Creator UserID",
    # )
    # creator: Optional[User] = Field(None,
    #     description="Post Creator User Model",
    #     # No need to define foreign_key and local_key as they assume id and creator_id
    #     #relation=BelongsTo('uvicore.auth.models.user.User', foreign_key='id', local_key='creator_id'),
    #     relation=BelongsTo('uvicore.auth.models.user.User'),
    # )
        # User model would have one-to-many (user has many posts)
        # posts: Optional[List[Post]] = Field(None,
        #     description="Users Posts Model",
        #     relation=HasMany('xx_vendor.xx_appname.models.post.Post', foreign_key='creator_id')
        # )



    # --------------------------------------------------------------------------
    # Example: One-To-Many Inverse (One Post has One Format)
    # But using key instead of id, so we explicitly define
    # --------------------------------------------------------------------------
    # format_key: str = Field('format_key',
    #     description='Post Format Key',
    # )
    # format: Optional[Format] = Field(None,
    #     description='Post Format Model',
    #     read_only=True,
    #     relation=BelongsTo('xx_vendor.xx_appname.models.Format', foreign_key='key', local_key='format_key'),
    # )



    # --------------------------------------------------------------------------
    # Example: Many-To-Many via post_tags pivot table
    # --------------------------------------------------------------------------
    # tags: Optional[List[Tag]] = Field(None,
    #     description="Post Tags",
    #     relation=BelongsToMany('xx_vendor.xx_appname.models.tag.Tag', join_tablename='post_tags', left_key='post_id', right_key='tag_id'),
    # )



    # --------------------------------------------------------------------------
    # Example: Polymorphic One-To-One image
    # --------------------------------------------------------------------------
    # image: Optional[Image] = Field(None,
    #     description="Post Image",
    #     relation=MorphOne('xx_vendor.xx_appname.models.image.Image', polyfix='imageable')
    # )



    # --------------------------------------------------------------------------
    # Example: Polymorphic Many-To-Many Hashtags
    # --------------------------------------------------------------------------
    #hashtags: Optional[List[Hashtag]] = Field(None,
    #hashtags: Optional[List[str]] = Field(None,
        # description="Post Hashtags",
        # relation=MorphToMany(
        #     model='xx_vendor.xx_appname.models.hashtag.Hashtag',
        #     join_tablename='hashtaggables',
        #     polyfix='hashtaggable',
        #     right_key='hashtag_id',
        #     # Or as a list
        #     #dict_key='id',
        #     #dict_value='name',
        #     #list_value='name',
        # ),



    # --------------------------------------------------------------------------
    # Example: Polymorphic One-To-Many attributes
    # --------------------------------------------------------------------------
    # attributes: Optional[List[Attribute]] = Field(None,
    # #attributes: Optional[Dict] = Field(None,  # As plain dict with dict_key and dict_value
    #     description="Post Attributes",
    #     # As plain dict
    #     #relation=MorphMany('xx_vendor.xx_appname.models.attribute.Attribute', polyfix='attributable', dict_key='key', dict_value='value')

    #     # Or as model
    #     relation=MorphMany('xx_vendor.xx_appname.models.attribute.Attribute', polyfix='attributable')



    # --------------------------------------------------------------------------
    # Example: Callback which is run AFTER the model has been instantiated
    # so it has access to self.  Think if this like a computed property.
    # --------------------------------------------------------------------------
    # cb: str = Field(None,
    #     callback='cb_results'
    # )
    # def cb_results(self):
    #     return str(self.slug) + ' callback'



    # --------------------------------------------------------------------------
    # Example: Evaluation logic which is run BEFORE the model has been instantiated.
    # so it has access to the incomming data (row), but not to self.
    # --------------------------------------------------------------------------
    # name: str = Field(None,
    #     # Using a inline lambda
    #     evaluate=lambda row: row['data']['name'] if 'data' in row else row['name']

    #     # Using a separate function
    #     evaluate=set_name

    #     # Using a separate function with parameters
    #     evaluate=(set_name, 'Data')
    # )



    # --------------------------------------------------------------------------
    # Example: Hooks into the model lifecycle
    # --------------------------------------------------------------------------
    # async def _before_insert(self) -> None:
    #     """Hook fired before record is inserted (new records only)"""
    #     await super()._before_insert()

    # async def _after_insert(self) -> None:
    #     """Hook fired after record is inserted (new records only)"""
    #     await super()._after_insert()

    # async def _before_save(self) -> None:
    #     """Hook fired before record is saved (inserted or updated)"""
    #     await super()._before_save()

    #     # Example: Convert password to hash if is plain text (works for first insert and updates)
    #     from uvicore.auth.support import password as pwd
    #     if self.password is not None and 'argon2' not in self.password:
    #         self.password = pwd.create(self.password)

    # async def _after_save(self) -> None:
    #     """Hook fired after record is saved (inserted or updated)"""
    #     await super()._after_save()

    # async def _before_delete(self) -> None:
    #     """Hook fired before record is deleted"""
    #     await super()._before_delete()

    # async def _after_delete(self) -> None:
    #     """Hook fired after record is deleted"""
    #     await super()._after_delete()



# ------------------------------------------------------------------------------
# Example: If models relate to each other, solve circular dependencies
# by importing some at the bottom and using update_forward_refs()
# ------------------------------------------------------------------------------
# Update forward refs to circumvent circular dependencies
# from uvicore.auth.models.user import User  # isort:skip
# from xx_vendor.xx_appname.models.post_info import PostInfo  # isort:skip
# from xx_vendor.xx_appname.models.tags import Tags  # isort:skip
# xx_ModelName.update_forward_refs()



from __future__ import annotations
import uvicore
from uvicore.typing import Optional
from uvicore.support.dumper import dd, dump
from uvicore.orm import Model, ModelMetaclass, Field
from uvicore.auth.database.tables import permissions as table


@uvicore.model()
class Permission(Model['Permission'], metaclass=ModelMetaclass):
    """Auth Permission Model"""

    __tableclass__ = table.Permissions

    id: Optional[int] = Field('id',
        primary=True,
        description='Permission ID',
    )

    entity: Optional[str] = Field('entity',
        description='Permission Entity',
    )

    name: str = Field('name',
        description='Permission Name',
    )



# Uvicore custom
from .authentication import Authentication

# Starlette passthrough via class proxy
from starlette.middleware.base import BaseHTTPMiddleware as _Base
from starlette.middleware.cors import CORSMiddleware as _CORS
from starlette.middleware.gzip import GZipMiddleware as _Gzip
from starlette.middleware.httpsredirect import HTTPSRedirectMiddleware as _HTTPSRedirect
from starlette.middleware.sessions import SessionMiddleware as _Session
from starlette.middleware.trustedhost import TrustedHostMiddleware as _TrustedHost
from starlette.middleware.wsgi import WSGIMiddleware as _WSGI


class Middleware(_Base):
    pass

class CORS(_CORS):
    pass

class Gzip(_Gzip):
    pass

class HTTPSRedirect(_HTTPSRedirect):
    pass

class Session(_Session):
    pass

class TrustedHost(_TrustedHost):
    pass

class WSGI(_WSGI):
    pass



from __future__ import annotations
import uvicore
from uvicore.typing import Optional, Dict, Union, List
from uvicore.auth.support import password as pwd
from uvicore.support.dumper import dd, dump
from uvicore.auth.models import Group, Role
from uvicore.auth.database.tables import users as table
from uvicore.orm import Model, ModelMetaclass, Field, BelongsTo, BelongsToMany
from uvicore.support.hash import sha1
from datetime import datetime


@uvicore.model()
class User(Model['User'], metaclass=ModelMetaclass):
    """Auth User Model"""

    # Database connection and table information
    __tableclass__ = table.Users

    id: Optional[int] = Field('id',
        primary=True,
        description='User Primary ID',
        #sortable=True,
        #searchable=True,
    )

    uuid: Optional[str] = Field('uuid',
        description='Custom Alternate UUID',
    )

    username: str = Field('username',
        description='User Login Username',
    )

    email: str = Field('email',
        description='User Email',
    )

    first_name: str = Field('first_name',
        description='User First Name',
    )

    last_name: str = Field('last_name',
        description='User Last Name',
    )

    title: Optional[str] = Field('title',
        description='User Title',
    )

    avatar_url: Optional[str] = Field('avatar_url',
        description='User Avatar URL',
    )

    password: Optional[str] = Field('password',
        description='User Last Name',
        read_only=False,
        write_only=True,
    )

    disabled: Optional[bool] = Field('disabled',
        description='User Disabled',
        default=False,
    )

    creator_id: int = Field('creator_id',
        description="User Creator UserID",
    )

    # One-To-Many Inverse (One Post has One Creator)
    creator: Optional[User] = Field(None,
        description="Post Creator User Model",
        #relation=BelongsTo('uvicore.auth.models.user.User', 'id', 'creator_id'),
        relation=BelongsTo('uvicore.auth.models.user.User'),
    )

    created_at: Optional[datetime] = Field('created_at',
        description='Created at Datetime',
        read_only=True,
    )

    updated_at: Optional[datetime] = Field('updated_at',
        description='Updated at Datetime',
        read_only=True,
    )

    login_at: Optional[datetime] = Field('login_at',
        description='Last Login Datetime',
    )

    # Many-To-Many via user_groups pivot table
    groups: Optional[List[Group]] = Field(None,
        description="User Groups",
        relation=BelongsToMany('uvicore.auth.models.group.Group', join_tablename='user_groups', left_key='user_id', right_key='group_id'),
    )

    # Many-To-Many via user_groups pivot table
    roles: Optional[List[Role]] = Field(None,
        description="User Roles",
        relation=BelongsToMany('uvicore.auth.models.role.Role', join_tablename='user_roles', left_key='user_id', right_key='role_id'),
    )

    async def _before_save(self) -> None:
        """Hook fired before record is saved (inserted or updated)"""
        await super()._before_save()

        # Convert password to hash if is plain text (works for first insert and updates)
        if self.password is not None and 'argon2' not in self.password:
            self.password = pwd.create(self.password)


# Required because I reference myself (creator_id)
User.update_forward_refs()



import uvicore
from uvicore.support import module
from uvicore.support.dumper import dump, dd
from uvicore.http.request import HTTPConnection
from uvicore.contracts import UserInfo, UserProvider
from uvicore.typing import Dict, Optional, List, Tuple
from uvicore.contracts import Logger as LoggerInterface
from uvicore.contracts import Authenticator as AuthenticatorInterface


@uvicore.service()
class Authenticator(AuthenticatorInterface):
    """Base authenticator class"""

    @property
    def log(self) -> LoggerInterface:
        return uvicore.log.name('uvicore.auth')

    def __init__(self, config: Dict):
        self.config = config

    async def retrieve_user(self, username: str, password: str, provider: Dict, request: HTTPConnection, **kwargs) -> Optional[UserInfo]:
        """Retrieve user from User Provider backend"""

        # Import user provider defined in auth config
        user_provider: UserProvider = module.load(provider.module).object()

        # Get user from user provider and validate password.  User will be Anonymous
        # if user not found, disabled or validation failed
        user = await user_provider.retrieve_by_credentials(
            # Require parameters
            username=username,
            password=password,
            request=request,

            # Pass in options from auth config
            **provider.options,

            # Pass in options from the calling authenticator
            **kwargs,
        )

        # Do not throw error if no user or not validated here.  We let the middleware handle that
        return user

    async def create_user(self, provider: Dict, request: HTTPConnection, **kwargs):
        # Import user provider defined in auth config
        user_provider: UserProvider = module.load(provider.module).object()

        # Create user from user provider
        # Returned user is actual backend user, NOT Auth User object
        user = await user_provider.create_user(request, **kwargs)
        return user

    async def sync_user(self, provider: Dict, request: HTTPConnection, **kwargs):
        # Import user provider defined in auth config
        user_provider: UserProvider = module.load(provider.module).object()

        # Create user from user provider
        # Returned user is actual backend user, NOT Auth User object
        user = await user_provider.sync_user(request, **kwargs)
        return user

    def auth_header(self, request) -> Tuple[str, str, str]:
        """Extract authorization header parts"""
        authorization = request.headers.get('Authorization')
        if not authorization: return (authorization, '', '')
        # Partition is a bit more performant that split
        scheme, _, param = authorization.partition(' ')
        return authorization, scheme.lower(), param



import inspect
from uvicore.typing import Any, Callable, List, Type, TypeVar, Union, get_type_hints
from fastapi import Depends
from uvicore.http.routing.web_router import WebRouter
from pydantic.typing import is_classvar
from starlette.routing import Route, WebSocketRoute
import uvicore
from uvicore.support.dumper import dump, dd

T = TypeVar("T")

CBV_CLASS_KEY = "__cbv_class__"



def controller(router: WebRouter):
    def decorator(cls):

        # cls = the decorated class = app1.http.controllers.home2.Home2

        #dd(inspect.signature(cls.__init__))

        dd(get_type_hints(cls))





        return cls
    return decorator














def WebController(router: WebRouter) -> Callable[[Type[T]], Type[T]]:
    """
    This function returns a decorator that converts the decorated into a class-based view for the provided router.
    Any methods of the decorated class that are decorated as endpoints using the router provided to this function
    will become endpoints in the router. The first positional argument to the methods (typically `self`)
    will be populated with an instance created using FastAPI's dependency-injection.
    For more detail, review the documentation at
    https://fastapi-utils.davidmontague.xyz/user-guide/class-based-views/#the-cbv-decorator
    """

    def decorator(cls: Type[T]) -> Type[T]:
        return _cbv(router, cls)

    return decorator


def _cbv(router: WebRouter, cls: Type[T]) -> Type[T]:
    """
    Replaces any methods of the provided class `cls` that are endpoints of routes in `router` with updated
    function calls that will properly inject an instance of `cls`.
    """

    # Runs at END of controller, so "router" is my router in controller
    # with /about route and all
    #dd(router)
    # for route in router.routes:
    #     dump(route.path)
    # dd('done')

    # Adds any class decorated arguments to class attributes
    _init_cbv(cls)

    # New web router to replace "router" as defined in the controller
    cbv_router = WebRouter(router.app, router.package, router.prefix, router.name)

    # An array of tuples of each class method including the new __init__ from _init_cbv(cls) above
    function_members = inspect.getmembers(cls, inspect.isfunction)
    #dd(function_members)

    # A set of just the controller methods includeing new __init__
    functions_set = set(func for _, func in function_members)
    #dd(functions_set)

    # List of Route objects in class defined router
    # Basically pulling out each route from my original class
    # cbv_routes = [
    #     route
    #     for route in router.routes
    #     if isinstance(route, (Route, WebSocketRoute)) and route.endpoint in functions_set
    # ]
    cbv_routes = router.routes
    #dd(cbv_routes)

    # Now remove each route from the controller defined router
    # And ADD them into this new router
    for route in cbv_routes.values():
        router.routes.remove(route)
        _update_cbv_route_endpoint_signature(cls, route)
        cbv_router.routes.append(route)

    # Now add the routes BACK to the controller routes but this
    # time they have the customized route signature
    #router.include_router(cbv_router)
    return cls


def _init_cbv(cls: Type[Any]) -> None:
    """
    Idempotently modifies the provided `cls`, performing the following modifications:
    * The `__init__` function is updated to set any class-annotated dependencies as instance attributes
    * The `__signature__` attribute is updated to indicate to FastAPI what arguments should be passed to the initializer
    """
    if getattr(cls, CBV_CLASS_KEY, False):  # pragma: no cover
        return  # Already initialized
    old_init: Callable[..., Any] = cls.__init__
    old_signature = inspect.signature(old_init)
    old_parameters = list(old_signature.parameters.values())[1:]  # drop `self` parameter
    new_parameters = [
        x for x in old_parameters if x.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
    ]
    dependency_names: List[str] = []
    for name, hint in get_type_hints(cls).items():
        if is_classvar(hint):
            continue
        parameter_kwargs = {"default": getattr(cls, name, Ellipsis)}
        dependency_names.append(name)
        new_parameters.append(
            inspect.Parameter(name=name, kind=inspect.Parameter.KEYWORD_ONLY, annotation=hint, **parameter_kwargs)
        )
    new_signature = old_signature.replace(parameters=new_parameters)

    def new_init(self: Any, *args: Any, **kwargs: Any) -> None:
        for dep_name in dependency_names:
            dep_value = kwargs.pop(dep_name)
            setattr(self, dep_name, dep_value)
        #old_init(self, *args, **kwargs)
        old_init(self)

    setattr(cls, "__signature__", new_signature)
    setattr(cls, "__init__", new_init)
    setattr(cls, CBV_CLASS_KEY, True)


def _update_cbv_route_endpoint_signature(cls: Type[Any], route: Union[Route, WebSocketRoute]) -> None:
    """
    Fixes the endpoint signature for a cbv route to ensure FastAPI performs dependency injection properly.
    """
    old_endpoint = route.endpoint
    #dd(old_endpoint)

    old_signature = inspect.signature(old_endpoint)
    #dd(old_signature)

    old_parameters: List[inspect.Parameter] = list(old_signature.parameters.values())
    #dd(old_parameters)

    old_first_parameter = old_parameters[0]
    new_first_parameter = old_first_parameter.replace(default=Depends(cls))
    new_parameters = [new_first_parameter] + [
        parameter.replace(kind=inspect.Parameter.KEYWORD_ONLY) for parameter in old_parameters[1:]
    ]
    # new_parameters = [
    #     parameter.replace(kind=inspect.Parameter.KEYWORD_ONLY) for parameter in old_parameters[1:]
    # ]
    #dd(new_parameters)
    new_signature = old_signature.replace(parameters=new_parameters)
    setattr(route.endpoint, "__signature__", new_signature)



# type: ignore
from typing import *
from .dictionary import Dict, OrderedDict
try:
    from starlette.types import Scope, Message, Receive, Send, ASGIApp
except:
    Scope = None
    Message = None
    Receive = None
    Send = None
    ASGIApp = None

# Decorator Type Helper
# def handle() -> Callable[[Decorator], Decorator]:
#     def decorator(func: Decorator) -> Decorator
#         return func
#     return decorator
Decorator = TypeVar("Decorator", bound=Callable[..., Any])

# from __future__ import annotations
# from typing import *
# import copy
# #from uvicore.typing.dictionary import _SuperDict
# from collections import OrderedDict as _OrderedDict
# from uvicore.support.dumper import dump, dd
# from uvicore.support.printer import pretty_call, register_pretty







# class _DictOLD:
#     def __getattr__(self, key):
#         """Getting a value by dot notation, pull from dictionary"""
#         ret = self.get(key)

#         # If key not in self dict, throw error
#         # This also hits if you do hasattr(x, y)
#         # NO, or else I can't do mydict.services or None of key doesn't exist
#         # This means I cannot use hasattr(x, y)
#         #if key not in self:
#         #    raise AttributeError()

#         # If key does exist but is None AND starts with __
#         if not ret and key.startswith("__"):
#             raise AttributeError()

#         # Key exists, even None, return value
#         return ret

#     def __setattr__(self, key, value):
#         self[key] = value

#     def __getstate__(self):
#         return self

#     def __setstate__(self, d):
#         self.update(d)

#     def copy(self):
#         """Create a clone copy of this dict"""
#         return self.__class__(dict(self).copy())

#     def update(self, d):
#         """Extend a dictionary with another dictionary"""
#         super(self.__class__, self).update(d)
#         return self

#     def merge(self, d):
#         """Deep merge self with this new dictionary"""
#         self.update(_deep_merge(d, self))
#         return self

#     def defaults(self, d):
#         """Provide defaults, essentially a reverse merge"""
#         self.update(_deep_merge(self, d))
#         return self

#     def extend(self, d):
#         """Alias for update"""
#         return self.update(d)

#     def clone(self):
#         """Alias of copy"""
#         return self.copy()

#     def hasattr(self, key):
#         return key in self



